// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_color.h
 *  \brief      MGL 色情報
 *  \date       Since: November 29, 2020. 15:17:59 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_COLOR_H_1606630679
#define INCGUARD_MGL_COLOR_H_1606630679

#include <cstdint>

namespace MGL
{
//! 色情報
struct Color
{
    float red;      //!< 赤成分
    float green;    //!< 緑成分
    float blue;     //!< 青成分
    float alpha;    //!< アルファ値
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief  ゼロ初期化
     */
    /* ------------------------------------------------------------------------- */
    constexpr Color() noexcept
        : red(0.0f)
        , green(0.0f)
        , blue(0.0f)
        , alpha(0.0f)
    {}
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      色を指定して初期化
     *  \param[in]  inRed   赤
     *  \param[in]  inGreen 緑
     *  \param[in]  inBlue  青
     *  \param[in]  inAlpha アルファ値
     */
    /* ------------------------------------------------------------------------- */
    constexpr Color(float inRed, float inGreen, float inBlue, float inAlpha = 1.0f) noexcept
        : red(inRed)
        , green(inGreen)
        , blue(inBlue)
        , alpha(inAlpha)
    {}
};


//! 各8bit RGBAの色情報
struct Color8bit
{
    uint8_t red;        //!< 赤成分
    uint8_t green;      //!< 緑成分
    uint8_t blue;       //!< 青成分
    uint8_t alpha;      //!< アルファ値
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief  ゼロ初期化
     */
    /* ------------------------------------------------------------------------- */
    constexpr Color8bit() noexcept
        : red(0)
        , green(0)
        , blue(0)
        , alpha(0)
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      色を指定して初期化
     *  \param[in]  inRed   赤
     *  \param[in]  inGreen 緑
     *  \param[in]  inBlue  青
     *  \param[in]  inAlpha アルファ値
     */
    /* ------------------------------------------------------------------------- */
    constexpr Color8bit(uint8_t inRed, uint8_t inGreen, uint8_t inBlue, uint8_t inAlpha = 255) noexcept
        : red(inRed)
        , green(inGreen)
        , blue(inBlue)
        , alpha(inAlpha)
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      32bit RGBA値で初期化
     *  \param[in]  rgba    RGBA値
     */
    /* ------------------------------------------------------------------------- */
    constexpr Color8bit(uint32_t rgba) noexcept
        : red(static_cast<uint8_t>(rgba >> 24))
        , green(static_cast<uint8_t>((rgba >> 16) & 0xFF))
        , blue((static_cast<uint8_t>(rgba >> 8) & 0xFF))
        , alpha(static_cast<uint8_t>(rgba & 0xFF))
    {}
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief  floatの色情報にキャスト
     */
    /* ------------------------------------------------------------------------- */
    constexpr operator Color() const noexcept
    {
        Color color;
        color.red = static_cast<float>(red) / 255.0f;
        color.green = static_cast<float>(green) / 255.0f;
        color.blue = static_cast<float>(blue) / 255.0f;
        color.alpha = static_cast<float>(alpha) / 255.0f;
        return color;
    }
};


constexpr Color kColorRed    = Color(1.0f, 0.0f, 0.0f, 1.0f);    //!< 赤
constexpr Color kColorGreen  = Color(0.0f, 1.0f, 0.0f, 1.0f);    //!< 緑
constexpr Color kColorBlue   = Color(0.0f, 0.0f, 1.0f, 1.0f);    //!< 青
constexpr Color kColorBlack  = Color(0.0f, 0.0f, 0.0f, 1.0f);    //!< 黒
constexpr Color kColorWhite  = Color(1.0f, 1.0f, 1.0f, 1.0f);    //!< 白

}   // namespace MGL

#endif	// INCGUARD_MGL_COLOR_H_1606630679

// vim: et ts=4 sw=4 sts=4
