// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_file_delegate_posix.h
 *  \brief      MGL ファイルデリゲート POSIX準拠
 *  \date       Since: January 20, 2021. 11:35:45 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_FILE_DELEGATE_POSIX_H_1611110145
#define INCGUARD_MGL_FILE_DELEGATE_POSIX_H_1611110145

#include <mgl/mgl_environment.h>
#if defined(MGL_FILE_DELEGATE_ENABLE_POSIX)

#include <mgl/file/mgl_file_delegate.h>
#include <stdio.h>

namespace MGL::File
{
//! POSIX準拠のファイルデリゲート
class POSIXDelegate : public Delegate
{
public:
    //! このデリゲートを表すデリゲートキー
    inline static constexpr DelegateKey kDelegateKey = MakeDelegateKey("MGL-POSIX");
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      このデリゲートがシステム標準のファイルを扱うものかを取得
     *  \retval     true    システム標準のファイルを扱う
     *  \retval     false   システム標準ではないファイルを扱う
     */
    /* ------------------------------------------------------------------------- */
    virtual bool IsManagedSystemNativeFile() const noexcept override
    {
        return true;    // 扱う
    }
        
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      このデリゲートが書き込みに対応しているかを取得
     *  \retval     true    対応している
     *  \retval     false   対応していない
     */
    /* ------------------------------------------------------------------------- */
    virtual bool IsWritable() const noexcept override
    {
        return true;    // 対応している
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      マウント時の処理
     *  \param[out] mountWork   マウント毎に使用するワークの共有ポインタ
     *  \param[in]  path        マウントパス
     *  \param[in]  accessType  アクセスタイプ
     *  \return     エラー発生時にいずれかのエラーが設定される
     */
    /* ------------------------------------------------------------------------- */
    virtual Result Mount([[maybe_unused]] SharedMountWork &mountWork, [[maybe_unused]] const char *path, [[maybe_unused]] MountAccessType accessType) noexcept override
    {
        return Result::Succeeded();     // このデリゲートはマウントワークを使用しないので何もせずに成功を返す
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      マウント解除時の処理
     *  \param[out] mountWork   マウント毎に使用するワークの共有ポインタ
     */
    /* ------------------------------------------------------------------------- */
    virtual void Unmount([[maybe_unused]] SharedMountWork &mountWork) noexcept override
    {
        return;
    }
    
    virtual AccessWorkPtr Open(Result &result, SharedMountWork &mountWork, const char *path, OpenMode mode) noexcept override;
    virtual Result Close(AccessWork *work) noexcept override;
    virtual size_t Read(AccessWork *work, Result &result, void *buffer, size_t size) noexcept override;
    virtual size_t Write(AccessWork *work, Result &result, const void *buffer, size_t size) noexcept override;
    virtual size_t Seek(AccessWork *work, Result &result, SeekType seekType, int32_t offset) noexcept override;
    virtual size_t GetOffset(AccessWork *work, Result &result) const noexcept override;
    virtual bool IsEOF(AccessWork *work, Result &result) const noexcept override;
    virtual size_t GetSize(AccessWork *work, Result &result) const noexcept override;
    virtual size_t GetSize(MountWork *mountWork, Result &result, const char *path) const noexcept override;
    virtual Result MakeDirectory(MountWork *mountWork, const char *path) noexcept override;
    virtual Result Move(MountWork *mountWork, const char *sourcePath, const char *destPath) noexcept override;
    virtual Result Remove(MountWork *mountWork, const char *path) noexcept override;
    virtual bool Exists(MountWork *mountWork, Result &result, const char *path) noexcept override;

    //! このデリゲート用の作業用構造体
    struct POSIXWork : public AccessWork
    {
        FILE *fp;       //!< ファイルポインタ
        size_t size;    //!< ファイルサイズ
        
        //! 初期化用コンストラクタ
        POSIXWork(OpenMode mode) noexcept
            : AccessWork(mode, kDelegateKey)
            , fp(nullptr)
            , size(0)
        {}        
    };

protected:
    POSIXWork *GetWork(AccessWork *work, OpenMode mode = OpenMode::None) const noexcept;
    const char *GetModeString(OpenMode mode) const noexcept;
    Error GetError(int errorNumber) const noexcept;
};
}   // namespace MGL::File

#endif  // MGL_FILE_DELEGATE_ENABLE_POSIX
#endif	// INCGUARD_MGL_FILE_DELEGATE_POSIX_H_1611110145

// vim: et ts=4 sw=4 sts=4
