// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_file_accessor.h
 *  \brief      MGL ファイルアクセサ
 *  \date       Since: January 20, 2021. 11:21:53 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_FILE_ACCESSOR_H_1611109313
#define INCGUARD_MGL_FILE_ACCESSOR_H_1611109313

#include <mgl/file/mgl_file_access_work.h>
#include <mgl/file/mgl_file_mounter.h>
#include <mgl/mgl_environment.h>
#include <mgl/stl/mgl_stl_containers.h>

namespace MGL::File::Accessor
{
AccessWorkPtr Open(Result &result, const char *path, OpenMode mode) noexcept;
Result Close(AccessWorkPtr &work) noexcept;

size_t Read(AccessWorkPtr &work, Result &result, void *buffer, size_t size) noexcept;
size_t Write(AccessWorkPtr &work, Result &result, const void *buffer, size_t size) noexcept;
size_t Seek(AccessWorkPtr &work, Result &result, SeekType seekType, int32_t offset) noexcept;
size_t GetOffset(AccessWorkPtr &work, Result &result) noexcept;
bool IsEOF(AccessWorkPtr &work, Result &result) noexcept;
size_t GetSize(AccessWorkPtr &work, Result &result) noexcept;

size_t GetSize(Result &result, const char *path) noexcept;
Result MakeDirectory(const char *path) noexcept;
Result Move(const char *sourcePath, const char *destPath) noexcept;
Result Remove(const char *path) noexcept;
Result Copy(const char *sourcePath, const char *destPath) noexcept;
bool Exists(Result &result, const char *path) noexcept;
bool IsSystemNativeFile(Result &result, const char *path) noexcept;

Result Mount(const char *mountName, const char *path, MountAccessType accessType, DelegateKey delegateKey = Mounter::kDefaultDelegateKey) noexcept;
Result Unmount(const char *mountName) noexcept;
Result Remount(const char *mountName, const char *path, MountAccessType accessType, DelegateKey delegateKey = Mounter::kDefaultDelegateKey) noexcept;

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      マウントされているかを取得
 *  \param[in]  mountName   マウント名
 *  \retval     true        マウントされている
 *  \retval     false       マウントされていない
 */
/* ------------------------------------------------------------------------- */
inline bool IsMounted(const char *mountName) noexcept
{
    return Mounter::GetInstance().IsMounted(mountName);
}

STL::string GetSystemNativePath(Result &result, const char *path) noexcept;

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      デリゲートの追加
 *  \tparam     DelegateClass   追加するデリゲートクラス
 *  \param[in]  key             デリゲートキー
 *  \param[in]  args            デリゲートクラスのコンストラクタに渡す引数
 *  \return     失敗時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
template <class DelegateClass, class... Args>
constexpr Result AddDelegate(DelegateKey key, Args... args) noexcept
{
    return Mounter::GetInstance().AddDelegate(key, STL::make_shared<DelegateClass>(args...));
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      デリゲートの削除
 *  \param[in]  key     削除するデリゲートキー
 *  \return     失敗時にはいずれかのエラーが設定される．存在しないデリゲートキーに対しては何もしない
 */
/* ------------------------------------------------------------------------- */
inline Result RemoveDelegate(DelegateKey key) noexcept
{
    return Mounter::GetInstance().RemoveDelegate(key);
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      デフォルトのデリゲートを設定
 *  \param[in]  key     デフォルトのデリゲートキー
 */
/* ------------------------------------------------------------------------- */
inline void SetDefaultDelegate(DelegateKey key) noexcept
{
    Mounter::GetInstance().SetDefaultDelegate(key);
}
}   // namespace MGL::File

#endif	// INCGUARD_MGL_FILE_ACCESSOR_H_1611109313

// vim: et ts=4 sw=4 sts=4
