// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_file_mounter.h
 *  \brief      MGL ファイルマウンタ
 *  \date       Since: January 20, 2021. 10:14:25 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_FILE_MOUNTER_H_1611105265
#define INCGUARD_MGL_FILE_MOUNTER_H_1611105265

#include <mgl/common/mgl_singleton.h>
#include <mgl/file/mgl_file_delegate.h>
#include <mgl/stl/mgl_stl_containers.h>
#include <mgl/stl/mgl_stl_memory.h>
#include <mgl/stl/mgl_stl_string.h>
#include <mutex>

namespace MGL::File
{
//! ファイルマウンタ
class Mounter final : public SharedSingleton<Mounter>
{
public:
    //! マウント情報
    struct MountProperty
    {
        STL::string name;               //!< マウント名
        STL::string path;               //!< マウントパス
        MountAccessType accessType;     //!< アクセスタイプ
        SharedDelegate delegate;        //!< ファイルデリゲート
        SharedMountWork mountWork;      //!< マウント毎のワーク
        
        //! 初期化用コンストラクタ
        MountProperty() noexcept
            : name()
            , path()
            , accessType(MountAccessType::ReadOnly)
            , delegate(nullptr)
            , mountWork(nullptr)
        {}
    };
    
    //! マウント情報の共有ポインタ
    using SharedMountProperty = std::shared_ptr<MountProperty>;

    //! デフォルトを表す予約デリゲートキー
    static constexpr DelegateKey kDefaultDelegateKey = MakeDelegateKey("MGL-Default");

    static STL::unique_ptr<Mounter> &GetInstanceRef() noexcept;
    
    Mounter() noexcept;
    
    Result Mount(const char *mountName, const char *path, MountAccessType accessType, DelegateKey delegateKey = kDefaultDelegateKey) noexcept;
    Result Unmount(const char *mountName) noexcept;
    SharedMountProperty Get(const char *mountName) noexcept;
    bool IsMounted(const char *mountName) noexcept;
    
    Result AddDelegate(DelegateKey key, SharedDelegate delegate) noexcept;
    Result RemoveDelegate(DelegateKey key) noexcept;
    SharedDelegate GetDelegate(DelegateKey key = kDefaultDelegateKey) noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      デフォルトのデリゲートを設定
     *  \param[in]  key     デフォルトのデリゲートキー
     */
    /* ------------------------------------------------------------------------- */
    void SetDefaultDelegate(DelegateKey key) noexcept
    {
        _defaultDelegateKey = key;
    }
    
private:
    STL::unordered_map<STL::string, std::shared_ptr<MountProperty>> _mountProperties;
    std::recursive_mutex _mountMutex;

    STL::unordered_map<DelegateKey, SharedDelegate> _delegates;
    std::mutex _delegateMutex;
    
    DelegateKey _defaultDelegateKey;
};
}   // namespace MGL::File
#endif	// INCGUARD_MGL_FILE_MOUNTER_H_1611105265

// vim: et ts=4 sw=4 sts=4
