// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_mouse_state.h
 *  \brief      MGL マウス入力状態定義
 *  \date       Since: December 28, 2020. 17:57:47 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_MOUSE_STATE_H_1609145867
#define INCGUARD_MGL_MOUSE_STATE_H_1609145867

#include <mgl/math/mgl_vector2.h>
#include <cstdint>
#include <mgl/common/mgl_bit.h>

namespace MGL::Input
{
//! マウスボタンのビットフラグ型
using MouseButtonFlags = uint32_t;

//! マウスボタン
enum class MouseButton : MouseButtonFlags
{
    None    = 0,                                //!< 入力なし
    Left    = MakeBit<MouseButtonFlags>(0),     //!< 左クリック
    Right   = MakeBit<MouseButtonFlags>(1),     //!< 右クリック
    Middle  = MakeBit<MouseButtonFlags>(2),     //!< 中央ボタン
    Other1  = MakeBit<MouseButtonFlags>(3),     //!< その他1（ボタン3）
    Other2  = MakeBit<MouseButtonFlags>(4),     //!< その他2（ボタン4）
    Other3  = MakeBit<MouseButtonFlags>(5),     //!< その他3（ボタン5）
    Other4  = MakeBit<MouseButtonFlags>(6),     //!< その他4（ボタン6）
    Other5  = MakeBit<MouseButtonFlags>(7),     //!< その他5（ボタン7）
    Other6  = MakeBit<MouseButtonFlags>(8),     //!< その他6（ボタン8）
    Other7  = MakeBit<MouseButtonFlags>(9),     //!< その他7（ボタン9）
    Other8  = MakeBit<MouseButtonFlags>(10),    //!< その他8（ボタン10）
};

//! マウスボタンのor演算の結果をフラグ型に変換するためのオペレータ（MouseButton | MouseButton）
constexpr MouseButtonFlags operator|(MouseButton lhs, MouseButton rhs) noexcept
{
    return static_cast<MouseButtonFlags>(lhs) | static_cast<MouseButtonFlags>(rhs);
}

//! マウスボタンのor演算の結果をフラグ型に変換するためのオペレータ（MouseButtonFlags | MouseButton）
constexpr MouseButtonFlags operator|(MouseButtonFlags lhs, MouseButton rhs) noexcept
{
    return static_cast<MouseButtonFlags>(lhs) | static_cast<MouseButtonFlags>(rhs);
}

//! マウスボタンのor演算の結果をフラグ型に変換するためのオペレータ （MouseButtonFlags |= MouseButton）
constexpr MouseButtonFlags operator|=(MouseButtonFlags &lhs, const MouseButton &rhs) noexcept
{
    return lhs | static_cast<MouseButtonFlags>(rhs);
}

//! マウスの全てのボタンを表す定数
constexpr MouseButtonFlags kMouseButtonAll = 0xFFFFFFFF;

//! マウスの状態
struct MouseState
{
    Vector2 position;           //!< ポインタの位置
    Vector2 prevPosition;       //!< 前回のポインタの位置
    Vector2 deltaMove;          //!< ポインタの移動量
    Vector2 uvPosition;         //!< ポインタの位置（UV座標）
    Vector2 uvPrevPosition;     //!< 前回のポインタの位置（UV座標）
    Vector2 uvDeltaMove;        //!< ポインタの移動量（UV座標）
    Vector2 deltaWheel;         //!< ホイール移動量（未実装）

    MouseButtonFlags buttonFlags;       //!< ボタンの状態
    MouseButtonFlags prevButtonFlags;   //!< 前回のボタンの状態

    //! 初期化用コンストラクタ
    constexpr MouseState() noexcept
        : position()
        , prevPosition()
        , deltaMove()
        , uvPosition()
        , uvPrevPosition()
        , uvDeltaMove()
        , deltaWheel()
        , buttonFlags(0)
        , prevButtonFlags(0)
    {}
};

//! カーソルモード
enum class CursorMode
{
    None,       //!< なし
    Pointer,    //!< ポインタ使用
    DeltaMove,  //!< 移動量のみ
};

//! 表示設定
enum class CursorVisibleMode
{
    Visible,        //!< 表示
    Invisible,      //!< 非表示
    AutoInvisible,  //!< 自動で非表示
};

}   // namespace MGL::Input

#endif	// INCGUARD_MGL_MOUSE_STATE_H_1609145867

// vim: et ts=4 sw=4 sts=4
