// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_memory_default_allocator.h
 *  \brief      MGL デフォルトのメモリアロケータ
 *  \date       Since: April 29, 2022. 20:24:04 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_MEMORY_DEFAULT_ALLOCATOR_H_1651231444
#define INCGUARD_MGL_MEMORY_DEFAULT_ALLOCATOR_H_1651231444

#include <mgl/memory/mgl_memory_allocator.h>

#include <mgl/memory/mgl_memory_fixed_size_allocator.h>
#include <stdio.h>
#include <stdlib.h>
#include <array>
#include <atomic>

namespace MGL::Memory
{
//! デフォルトのメモリアロケータ
class DefaultAllocator : public Allocator
{
public:
    //! デフォルトアロケータの設定
    struct Configuration
    {
        //! 固定サイズアロケータの最小サイズ（2のn乗の値である必要がある）
        size_t fixedSizeAllocatorMinimumBlockSize;
        
        //! 固定サイズアロケータのメモリプールのサイズ（最小サイズの8倍より大きくなければならない）
        size_t fixedSizeAllocatorMemoryPoolSize;
        
        //! 固定サイズアロケータが使用できない場合のアロケータの指定（nullptr で malloc()を使用）
        void *(*allocator)(size_t size);

        //! 固定サイズアロケータが使用できない場合のデアロケータの指定（nullptr で free()を使用）
        void (*deallocator)(void *buffer);
        
        //! コンストラクタ
        constexpr Configuration() noexcept
            : fixedSizeAllocatorMinimumBlockSize(128)               // 最小128バイト
            , fixedSizeAllocatorMemoryPoolSize(1 * 1024 * 1024)     // メモリプールは1MiB
            , allocator(nullptr)                                    // アロケータに malloc()を使用
            , deallocator(nullptr)                                  // デアロケータに free()を使用
        {
        }
    };

    DefaultAllocator() noexcept;

    //! このアロケータを表す値
    static constexpr AllocatorType kAllocatorType = MakeAllocatorType("MGL-DefaultAllocator");

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      このアロケータを表す値を取得
     *  \return     このアロケータを表す値
     *  \note       MakeAllocatorType()によって生成された値を推奨
     */
    /* ------------------------------------------------------------------------- */
    virtual AllocatorType GetType() const noexcept override
    {
        return kAllocatorType;
    }

    bool Initialize(const Configuration &config) noexcept;

    virtual bool Initialize() noexcept override;
    virtual bool IsInitialized() const noexcept override;
    [[nodiscard]] virtual void *Allocate(size_t size) noexcept override;
    virtual void Deallocate(void *buffer) noexcept override;

    virtual bool GetSizeInfo(size_t &dest, uint32_t key, uint32_t arg) noexcept override;

private:
    bool _isInitialized;
    
    // 固定メモリアロケータの数．アルゴリズムに関わるので変更不可．
    static constexpr size_t kFixedSizeAllocatorCount = 4;

    Configuration _config;
    std::array<FixedSizeAllocator, kFixedSizeAllocatorCount> _fixedSizeAllocatorArray;
    size_t _fixedSizeBlockMax;
    uint8_t _shiftBit;

    std::atomic<size_t> _systemUsedCount;
    std::atomic<size_t> _systemUsedSize;
};

}   // namespace MGL::Memory

#endif	// INCGUARD_MGL_MEMORY_DEFAULT_ALLOCATOR_H_1651231444

// vim: et ts=4 sw=4 sts=4
