// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_memory_fixed_size_allocator.h
 *  \brief      MGL 固定サイズアロケータ
 *  \date       Since: May 22, 2022. 18:20:51 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_MEMORY_FIXED_SIZE_ALLOCATOR_H_1653211251
#define INCGUARD_MGL_MEMORY_FIXED_SIZE_ALLOCATOR_H_1653211251

#include <forward_list>
#include <memory>
#include <mutex>

namespace MGL::Memory
{
//! 固定サイズアロケータクラス
class FixedSizeAllocator
{
public:
    //! メモリブロック情報
    struct Block
    {
        uintptr_t address;  //!< メモリブロックのアドレス
        Block *next;        //!< 次のメモリブロック
    };
    
    //! ヘッダ情報
    struct Header
    {
        FixedSizeAllocator *allocator;  //!< 使用したアロケータ

        union
        {
            Block *block;            //!< 使用したメモリブロック（固定サイズアロケータを使用した場合）
            size_t requestSize;      //!< 要求サイズ（固定サイズアロケータを使用しなかった場合）
        };
    };
    static_assert(sizeof(Header) <= 16);    // ヘッダサイズは16バイト以下でなければならない

    FixedSizeAllocator() noexcept;

    bool Initialize(size_t blockSize, size_t blockCount) noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      代替アロケータを設定
     *  \param[in]  altAllocator    設定する代替アロケータ
     */
    /* ------------------------------------------------------------------------- */
    constexpr void SetAltAllocator(FixedSizeAllocator *altAllocator) noexcept
    {
        _altAllocator = altAllocator;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ブロックサイズを取得
     *  \return     ブロックサイズ
     */
    /* ------------------------------------------------------------------------- */
    constexpr size_t GetBlockSize() const noexcept
    {
        return _blockSize;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      確保可能な最大数を取得
     *  \return     確保可能な最大数
     */
    /* ------------------------------------------------------------------------- */
    constexpr size_t GetCapacity() const noexcept
    {
        return _blockCount;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      利用可能なブロック数を取得
     *  \return     利用可能なブロック数
     */
    /* ------------------------------------------------------------------------- */
    size_t GetFreeCount() noexcept
    {
        std::lock_guard lock(_mutex);
        return _freeCount;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      利用中のブロック数を取得
     *  \return     利用中のブロック数
     */
    /* ------------------------------------------------------------------------- */
    size_t GetUsedCount() noexcept
    {
        return GetCapacity() - GetFreeCount();
    }

    void *Allocate(size_t size) noexcept;
    void Deallocate(void *buffer) noexcept;

private:
    size_t _blockSize;
    size_t _blockCount;
    std::unique_ptr<std::byte[]> _pool;
    size_t _poolSize;
    std::unique_ptr<Block[]> _blockArray;
    Block *_topBlock;
    size_t _freeCount;
    FixedSizeAllocator *_altAllocator;

    std::mutex _mutex;
};
}   // namespace MGL::Memory

#endif	// INCGUARD_MGL_MEMORY_FIXED_SIZE_ALLOCATOR_H_1653211251

// vim: et ts=4 sw=4 sts=4
