// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_win32_window.h
 *  \brief      MGL Win32用ウィンドウ
 *  \date       Since: March 1, 2021. 12:04:32 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_WIN32_WINDOW_H_1614567872
#define INCGUARD_MGL_WIN32_WINDOW_H_1614567872

#include <mgl/mgl_environment.h>
#if defined(MGL_TARGET_WIN32)

#include <Windows.h>
#include <mgl/common/mgl_singleton.h>
#include <mgl/event/mgl_event_handle.h>
#include <mgl/stl/mgl_stl_memory.h>

namespace MGL::Win32
{
//! Win32用ウィンドウクラス
class Window final : public MGL::SharedSingleton<Window>
{
public:

    //! ウィンドウ初期化用記述子
    struct Descriptor
    {
        int width;              //!< ウィンドウの幅
        int height;             //!< ウィンドウの高さ
        LPCWSTR windowTitle;    //!< ウィンドウタイトル
        LPCWSTR className;      //!< クラス名（nullptr指定でウィンドウタイトルを使用）
        HICON icon;             //!< ウィンドウアイコン
        HICON smallIcon;        //!< ウィンドウの小アイコン

        //! コンストラクタ
        constexpr Descriptor() noexcept
            : width(1280)
            , height(720)
            , windowTitle(L"MGL-App")
            , className(nullptr)
            , icon(nullptr)
            , smallIcon(nullptr)
        {
        }
    };

    static STL::unique_ptr<Window> &GetInstanceRef() noexcept;

    Window() noexcept;

    bool Initialize(HINSTANCE hInstance, int nCmdShow, WNDPROC windowProc, const Descriptor &descriptor) noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ウィンドウハンドラの取得
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR HWND GetWindowHandler() noexcept
    {
        return _hWnd;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ウィンドウの幅を取得
     *  \return     ウィンドウの幅
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR int GetWidth() const noexcept
    {
        return _width;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ウィンドウの高さを取得
     *  \return     ウィンドウの高さ
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR int GetHeight() const noexcept
    {
        return _height;
    }


    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      アプリケーションインスタンスのハンドルを取得
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR HINSTANCE GetAppInstanceHandle() const noexcept
    {
        return _hInstance;
    }

    void SetWindowTitle(const char *title) noexcept;
    bool Resize(int width, int height) noexcept;
    float GetLogicalScaleFactor() const noexcept;

    void Update() noexcept;

private:
    HINSTANCE _hInstance;
    HWND _hWnd;
    int _width;
    int _height;
    float _scaleFactor;
};
}   // namespace MGL::Win32

#endif  // MGL_TARGET_WIN32
#endif	// INCGUARD_MGL_WIN32_WINDOW_H_1614567872

// vim: et ts=4 sw=4 sts=4
