// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_font_resource.h
 *  \brief      MGL フォントリソース
 *  \date       Since: May 26, 2021. 14:59:28 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */
#ifndef INCGUARD_MGL_FONT_RESOURCE_H_1622008768
#define INCGUARD_MGL_FONT_RESOURCE_H_1622008768

#include <mgl/math/mgl_vector2.h>
#include <mgl/render/font/mgl_font_option.h>
#include <mgl/text/mgl_text_index_converter.h>



namespace MGL::Render
{
//! フォントリソースクラス
class FontResource
{
public:
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     */
    /* ------------------------------------------------------------------------- */
    constexpr FontResource(bool supportIndexedCharacter) noexcept
        : _indexConverter(*this)
        , _supportIndexedCharacter(supportIndexedCharacter)
    {
    }
    
    virtual ~FontResource() noexcept = default;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      文字の表示
     *  \param[in]  position    表示位置
     *  \param[in]  option      表示オプション
     *  \param[in]  text        表示する文字列
     *  \return     表示後の文字の位置
     */
    /* ------------------------------------------------------------------------- */
    virtual Vector2 Print(const Vector2 position, const FontOption &option, const char *text) noexcept = 0;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      インデックス化した文字を表示
     *  \param[in]  position        表示位置
     *  \param[in]  option          表示オプション
     *  \param[in]  indexedString   表示する文字列
     *  \return     表示後の文字の位置
     */
    /* ------------------------------------------------------------------------- */
    virtual Vector2 Print(const Vector2 position, [[maybe_unused]] const FontOption &option, [[maybe_unused]] const Text::IndexedCharacter *indexedString) noexcept
    {
        return position;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      文字をインデックス文字に変換
     *  \param[in]  character   変換元の文字（UTF-32）
     *  \return     インデックス文字．失敗時はMGL::Text::kIndexedCharacterInvalid
     */
    /* ------------------------------------------------------------------------- */
    virtual Text::IndexedCharacter ToIndexedCharacter([[maybe_unused]] char32_t character) const noexcept
    {
        return Text::kIndexedCharacterInvalid;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      インデックス文字を通常の文字に変換
     *  \param[in]  indexedCharacter    インデックス文字
     *  \return     インデックスに対応したUTF-32エンコードの文字．失敗時はNULL文字(0x00)
     */
    /* ------------------------------------------------------------------------- */
    virtual char32_t ToCharacter([[maybe_unused]] Text::IndexedCharacter indexedCharacter) const noexcept
    {
        return 0;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      このフォントリソースが有効かどうかを取得
     *  \retval     true    有効
     *  \retval     false   無効
     */
    /* ------------------------------------------------------------------------- */
    virtual bool IsValid() const noexcept = 0;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      このフォントリソースがインデックス文字をサポートしているかを取得
     *  \retval     true    サポートしている
     *  \retval     false   サポートしていない
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool CanConvertIndexedString() const noexcept
    {
        return _supportIndexedCharacter;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      インデックス文字列のコンバータを取得
     *  \return     インデックス文字列のコンバータ
     */
    /* ------------------------------------------------------------------------- */
    constexpr const Text::IndexConverter &GetIndexConverter() const noexcept
    {
        return _indexConverter;
    }
    
private:
    //! フォントリソースクラス用のインデックス文字列コンバータ
    class IndexConverter : public Text::IndexConverter
    {
    public:
        /* ------------------------------------------------------------------------- */
        /*!
         *  \brief      コンストラクタ
         */
        /* ------------------------------------------------------------------------- */
        constexpr IndexConverter(const FontResource &fontResource) noexcept
            : _resource(fontResource)
        {
        }
        
        /* ------------------------------------------------------------------------- */
        /*!
         *  \brief      文字をインデックス文字に変換
         *  \param[in]  character   変換元の文字（UTF-32）
         *  \return     インデックス文字
         */
        /* ------------------------------------------------------------------------- */
        virtual Text::IndexedCharacter ToIndexedCharacter(char32_t character) const noexcept override
        {
            return _resource.ToIndexedCharacter(character);
        }
        
        /* ------------------------------------------------------------------------- */
        /*!
          *  \brief      インデックス文字を通常の文字に変換
          *  \param[in]  indexedCharacter    インデックス文字
          *  \return     インデックスに対応したUTF-32エンコードの文字
          */
         /* ------------------------------------------------------------------------- */
        virtual char32_t ToCharacter(Text::IndexedCharacter indexedCharacter) const noexcept override
        {
            return _resource.ToCharacter(indexedCharacter);
        }

    private:
        const FontResource &_resource;
    };
    
    IndexConverter _indexConverter;
    bool _supportIndexedCharacter;
};

//! フォントリソースの共有ポインタ
using SharedFontResource = std::shared_ptr<FontResource>;

}   // namespace MGL::Render

#endif	// INCGUARD_MGL_FONT_RESOURCE_H_1622008768

// vim: et ts=4 sw=4 sts=4
