// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_system_locale.h
 *  \brief      MGL ロケール情報
 *  \date       Since: February 24, 2021. 2:07:07 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_SYSTEM_LOCALE_H_1614100027
#define INCGUARD_MGL_SYSTEM_LOCALE_H_1614100027

#include <mgl/system/mgl_system_module_set.h>
#include <mgl/event/mgl_event.h>
#include <mgl/mgl_environment.h>

namespace MGL::System
{
//! ロケール情報クラス
class Locale
{
public:
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     */
    /* ------------------------------------------------------------------------- */
    Locale() noexcept
        : _locale(ModuleSet::GetInstance().GetLocaleModule())
    {}
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      言語を取得
     *  \return     現在設定されている言語
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR Language GetLanguage() const noexcept
    {
        return _locale.GetLocaleInfo().language;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      デフォルトの言語を取得
     *  \return     デフォルトの言語
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR Language GetDefaultLanguage() const noexcept
    {
        return _locale.GetDefaultLocaleInfo().language;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      言語を設定
     *  \param[in]  language    設定する言語
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR void SetLanguage(Language language) noexcept
    {
        auto &currentLocaleInfo = _locale.GetLocaleInfo();
        if (currentLocaleInfo.language != language)
        {
            currentLocaleInfo.language = language;
            Event::Notify(Event::NotifyType::ChangeLanguage, &currentLocaleInfo.language);
        }
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      UTC基準の時差を取得
     *  \return     UTC基準の時差（秒）
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR int32_t GetTimeDifferent() const noexcept
    {
        return _locale.GetLocaleInfo().timeDifferent;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      デフォルトのUTC基準の時差を取得
     *  \return     デフォルトのUTC基準の時差（秒）
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR int32_t GetDefaultTimeDifferent() const noexcept
    {
        return _locale.GetDefaultLocaleInfo().timeDifferent;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      UTC基準の時差を設定
     *  \param[in]  timeDifferent   設定する時差
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR void SetTimeDifferent(int32_t timeDifferent) noexcept
    {
        _locale.GetLocaleInfo().timeDifferent = timeDifferent;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      桁区切りの間隔を取得
     *  \return     現在設定されている桁区切りの間隔
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR uint8_t GetGroupingCount() const noexcept
    {
        return _locale.GetLocaleInfo().groupingCount;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      デフォルトの桁区切りの間隔を取得
     *  \return     デフォルトの桁区切りの間隔
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR uint8_t GetDefaultGroupingCount() const noexcept
    {
        return _locale.GetDefaultLocaleInfo().groupingCount;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      桁区切りの間隔を設定
     *  \param[in]  groupingCount   桁区切りの間隔
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR void SetGroupingCount(uint8_t groupingCount) noexcept
    {
        _locale.GetLocaleInfo().groupingCount = groupingCount;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      数値の桁区切り記号を取得
     *  \return     現在設定されている桁区切り記号
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR const char *GetGroupingSeparator() const noexcept
    {
        return _locale.GetLocaleInfo().groupingSeparator.c_str();
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      デフォルトの数値の桁区切り記号を取得
     *  \return     デフォルトの桁区切り記号
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR const char *GetDefaultGroupingSeparator() const noexcept
    {
        return _locale.GetDefaultLocaleInfo().groupingSeparator.c_str();
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      数値の桁区切り記号を設定
     *  \param[in]  separator   区切り記号
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR void SetGroupingSeparator(const char *separator) noexcept
    {
        _locale.GetLocaleInfo().groupingSeparator = separator;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      小数点の区切り記号を取得
     *  \return     現在設定されている区切り記号
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR const char *GetDecimalSeparator() const noexcept
    {
        return _locale.GetLocaleInfo().decimalSeparator.c_str();
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      デフォルトの小数点の区切り記号を取得
     *  \return     デフォルトの小数点の区切り記号
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR const char *GetDefaultDecimalSeparator() const noexcept
    {
        return _locale.GetDefaultLocaleInfo().decimalSeparator.c_str();
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      小数点の区切り記号を設定
     *  \param[in]  separator   区切り記号
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR void SetDecimalSeparator(const char *separator) noexcept
    {
        _locale.GetLocaleInfo().decimalSeparator = separator;
    }

private:
    LocaleModuleInterface &_locale;
};
}   // namespace MGL::System

#endif	// INCGUARD_MGL_SYSTEM_LOCALE_H_1614100027

// vim: et ts=4 sw=4 sts=4
