// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_task.h
 *  \brief      MGL タスクシステム
 *  \date       Since: June 3, 2021. 11:46:50 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_TASK_H_1622688410
#define INCGUARD_MGL_TASK_H_1622688410

#include <mgl/mgl_environment.h>
#include <mgl/task/mgl_task_node_list.h>
#include <mgl/task/mgl_task_weak_node.h>
#include <mgl/task/mgl_task_default.h>

namespace MGL::Task
{
bool Initialize(size_t capacity, const ExecuteSettingArray &executeSettings = ExecuteSettingArray()) noexcept;

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      タスクの生成
 *  \tparam     TaskClass   生成するタスクのクラス
 *  \param[in]  args        コンストラクタに与える引数
 *  \return     生成されたタスクの弱参照ノード．失敗している場合は無効な弱参照ノード
 */
/* ------------------------------------------------------------------------- */
template <class TaskClass, class... Args>
constexpr WeakNode Create(Args... args) noexcept
{
    auto &nodeList = NodeList::GetInstance();
    if (!nodeList.HasFreeNode())
    {
        return WeakNode();
    }
    
    auto index = nodeList.Add(STL::make_unique<TaskClass>(args...));

    return WeakNode(index);
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      タスクの実行
 */
/* ------------------------------------------------------------------------- */
inline void Execute() noexcept
{
    NodeList::GetInstance().Execute();
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      全てのタスクに削除要求を発行
 *  \param[in]  resideLevel    常駐レベルがこの値以下のタスクを削除する
 *  \return     削除要求を発行した数
 */
/* ------------------------------------------------------------------------- */
inline size_t Kill(ResideLevel resideLevel = ResideLevel::NoResident) noexcept
{
    return NodeList::GetInstance().Kill(resideLevel);
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      指定したIDに対して削除要求を発行
 *  \param[in]  identifier      削除するID
 *  \param[in]  resideLevel     常駐レベルがこの値以下のタスクを削除する
 *  \return     削除要求を発行した数
 */
/* ------------------------------------------------------------------------- */
inline size_t Kill(Identifier identifier, ResideLevel resideLevel = ResideLevel::NoResident) noexcept
{
    return NodeList::GetInstance().Kill(identifier, resideLevel);
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      常駐プライオリティを無視して全てのタスクを削除
 *  \return     削除要求を発行した数
 */
/* ------------------------------------------------------------------------- */
inline size_t DestroyThemAll() noexcept
{
    return Kill(ResideLevel::Max);
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      現在のタスクの総数を取得
 *  \return     現在のタスクの総数
 */
/* ------------------------------------------------------------------------- */
inline size_t GetCount() noexcept
{
    return NodeList::GetInstance().GetCount();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      指定したIDのタスクの数を取得
 *  \return     指定したIDのタスクの数
 */
/* ------------------------------------------------------------------------- */
inline size_t GetCount(Identifier identifier) noexcept
{
    return NodeList::GetInstance().GetCount(identifier);
}

STL::vector<WeakNode> Find(Identifier identifier) noexcept;

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      タスクへのイベント通知
 *  \param[in]  eventIdentifier     イベントID
 *  \param[in]  argument            引数
 */
/* ------------------------------------------------------------------------- */
inline void NotifyEvent(uint32_t eventIdentifier, void *argument = nullptr) noexcept
{
    return NodeList::GetInstance().NotifyEvent(eventIdentifier, argument);
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      タスクへのイベント通知
 *  \param[in]  identifier          タスクID
 *  \param[in]  eventIdentifier     イベントID
 *  \param[in]  argument            引数
 */
/* ------------------------------------------------------------------------- */
inline void NotifyEvent(Identifier identifier, uint32_t eventIdentifier, void *argument = nullptr) noexcept
{
    return NodeList::GetInstance().NotifyEvent(identifier, eventIdentifier, argument);
}

}   // namespace MGL::Task

#endif	// INCGUARD_MGL_TASK_H_1622688410

// vim: et ts=4 sw=4 sts=4
