// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_task_node.h
 *  \brief      MGL タスクノード
 *  \date       Since: June 1, 2021. 16:46:33 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_TASK_NODE_H_1622533593
#define INCGUARD_MGL_TASK_NODE_H_1622533593

#include <mgl/task/mgl_task_defs.h>

namespace MGL::Task
{
//! タスクノード
class Node
{
public:
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     *  \param[in]  identify    タスクノードのID
     */
    /* ------------------------------------------------------------------------- */
    constexpr Node(Identifier identify) noexcept
        : _identify(identify)
        , _isRequestedKill(false)
        , _resideLevel(ResideLevel::NoResident)
    {}

    virtual ~Node() = default;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      このタスクノードのIDを取得
     *  \return     タスクノードのID
     */
    /* ------------------------------------------------------------------------- */
    constexpr Identifier GetIdentifier() const noexcept
    {
        return _identify;
    }    
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      削除要求
     */
    /* ------------------------------------------------------------------------- */
    constexpr void Kill() noexcept
    {
        _isRequestedKill = true;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      削除要求されているかを取得
     *  \retval     true    されている
     *  \retval     false   されていない
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool IsRequestedKill() const noexcept
    {
        return _isRequestedKill;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      常駐タスクの設定
     *  \param[in]  isEnabled   trueを指定すると常駐タスクになり，falseで解除する
     */
    /* ------------------------------------------------------------------------- */
    [[deprecated]] constexpr void SetResident(bool isEnabled) noexcept
    {
        _resideLevel = (isEnabled) ? ResideLevel::NoResident : ResideLevel::Middle;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      常駐レベルの設定
     *  \param[in]  resideLevel     設定する常駐レベル
     */
    /* ------------------------------------------------------------------------- */
    constexpr void SetResideLevel(ResideLevel resideLevel) noexcept
    {
        _resideLevel = resideLevel;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      常駐レベルの取得
     *  \return     現在の常駐レベル
     */
    /* ------------------------------------------------------------------------- */
    constexpr ResideLevel GetResideLevel() const noexcept
    {
        return _resideLevel;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      常駐タスクかどうかを取得
     *  \retval     true    常駐タスクである
     *  \retval     false   常駐タスクでない
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool IsResident() const noexcept
    {
        return _resideLevel > ResideLevel::NoResident;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      タスク生成時に呼び出される関数
     */
    /* ------------------------------------------------------------------------- */
    virtual void OnInitialize() noexcept
    {
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      タスク実行時に呼び出される関数
     *  \param[in]  stage   実行ステージ
     */
    /* ------------------------------------------------------------------------- */
    virtual void OnExecute(ExecuteStage stage) noexcept = 0;
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      タスクのイベント受信処理
     *  \param[in]  eventIdentifier     イベントID
     *  \param[in]  argument            引数
     */
    /* ------------------------------------------------------------------------- */
    virtual void OnReceiveTaskEvent(EventIdentifier eventIdentifier, void *argument) noexcept
    {
        (void)eventIdentifier;
        (void)argument;
    }

private:
    Identifier _identify;
    bool _isRequestedKill;
    ResideLevel _resideLevel;
};
}   // namespace MGL::Task

#endif	// INCGUARD_MGL_TASK_NODE_H_1622533593

// vim: et ts=4 sw=4 sts=4
