// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_task_node_list.h
 *  \brief      MGL タスクノードのリスト
 *  \date       Since: June 1, 2021. 17:06:07 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_TASK_NODE_LIST_H_1622534767
#define INCGUARD_MGL_TASK_NODE_LIST_H_1622534767

#include <mgl/common/mgl_singleton.h>
#include <mgl/task/mgl_task_node.h>
#include <mgl/stl/mgl_stl_containers.h>
#include <mgl/stl/mgl_stl_memory.h>

namespace MGL::Task
{
//! タスクノードのリスト
class NodeList final : public SharedSingleton<NodeList>
{
public:
    //! ノードの要素
    struct NodeElement
    {
        STL::unique_ptr<Node> node; //!< このノード
        NodeElement *next;          //!< 次のノード
        NodeElement *previous;      //!< 前のノード
        UniqueIdentifier uniqueId;  //!< ユニークID
        size_t arrayIndex;          //!< 配列上のインデックス
        bool isStandby;             //!< スタンバイフラグ

        //! コンストラクタ
        constexpr NodeElement() noexcept
            : node(nullptr)
            , next(nullptr)
            , previous(nullptr)
            , uniqueId(UniqueIdentifier::Invalid)
            , arrayIndex(0)
            , isStandby(true)
        {}
    };

    static STL::unique_ptr<NodeList> &GetInstanceRef() noexcept;
    
    NodeList() noexcept;

    bool Initialize(size_t capacity, const ExecuteSettingArray &settings) noexcept;
    
    size_t Add(STL::unique_ptr<Node> node) noexcept;

    void Execute() noexcept;

    size_t Kill(ResideLevel priority = ResideLevel::NoResident) noexcept;
    size_t Kill(Identifier identify, ResideLevel priority = ResideLevel::NoResident) noexcept;

    size_t GetCount() noexcept;
    size_t GetCount(Identifier identify) noexcept;

    const NodeElement *GetElement(size_t index) const noexcept;
    const NodeElement *GetElement(const Node *node) const noexcept;

    STL::vector<size_t> Find(Identifier identify) const noexcept;
    
    void NotifyEvent(EventIdentifier eventIdentifier, void *arg) noexcept;
    void NotifyEvent(Identifier identifier, EventIdentifier eventIdentifier, void *arg) noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      追加可能なノードの数を取得
     *  \return     追加可能なノードの数
     */
    /* ------------------------------------------------------------------------- */
    constexpr size_t GetCapacity() const noexcept
    {
        return _capacity;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      空きノードがあるかをチェック
     *  \retval     true    空きノードがある
     *  \retval     false   空きノードがない
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool HasFreeNode() const noexcept
    {
        return _freeTopElement.next != nullptr;
    }
        
private:
    void Remove(NodeElement *element) noexcept;

    size_t _capacity;
    STL::unique_ptr<NodeElement[]> _nodeElements;
    NodeElement _activeTopElement;
    NodeElement _freeTopElement;

    UniqueIdentifier _currentUniqueId;

    ExecuteSettingArray _executeSettings;
};
}   // namespace MGL::Task

#endif	// INCGUARD_MGL_TASK_NODE_LIST_H_1622534767

// vim: et ts=4 sw=4 sts=4
