// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_text_format_argument.h
 *  \brief      MGL テキストフォーマットの引数
 *  \date       Since: May 28, 2021. 22:25:20 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_TEXT_FORMAT_ARGUMENT_H_1622208320
#define INCGUARD_MGL_TEXT_FORMAT_ARGUMENT_H_1622208320

#include <mgl/text/mgl_text_defs.h>
#include <mgl/stl/mgl_stl_containers.h>
#include <mgl/stl/mgl_stl_string.h>

namespace MGL::Text
{
//! テキストフォーマットの引数
class FormatArgument
{
public:
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument() noexcept
        : _type(Type::Void)
        , _i(0)
        , _string()
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(unsigned long long)
     *  \param[in]  value   符号なしの値
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(unsigned long long value) noexcept
        : _type(Type::UnsignedInt)
        , _ui(value)
        , _string()
    {}
        
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(unsigned long)
     *  \param[in]  value   符号なしの値
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(unsigned long value) noexcept
        : FormatArgument(static_cast<unsigned long long>(value))
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(unsigned int)
     *  \param[in]  value   符号なしの値
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(unsigned int value) noexcept
        : FormatArgument(static_cast<unsigned long long>(value))
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(unsigned short)
     *  \param[in]  value   符号なしの値
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(unsigned short value) noexcept
        : FormatArgument(static_cast<unsigned long long>(value))
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(unsigned char)
     *  \param[in]  value   符号なしの値
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(unsigned char value) noexcept
        : FormatArgument(static_cast<unsigned long long>(value))
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(signed long long)
     *  \param[in]  value   符号付きの値
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(signed long long value) noexcept
        : _type(Type::Int)
        , _i(value)
        , _string()
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(signed long)
     *  \param[in]  value   符号付きの値
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(signed long value) noexcept
        : FormatArgument(static_cast<signed long long>(value))
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(signed int)
     *  \param[in]  value   符号付きの値
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(signed int value) noexcept
        : FormatArgument(static_cast<signed long long>(value))
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(signed short)
     *  \param[in]  value   符号付きの値
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(signed short value) noexcept
        : FormatArgument(static_cast<signed long long>(value))
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(signed char)
     *  \param[in]  value   符号付きの値
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(signed char value) noexcept
        : FormatArgument(static_cast<signed long long>(value))
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(float)
     *  \param[in]  value   浮動小数点数
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(float value) noexcept
        : _type(Type::Float)
        , _ui(0)
        , _string()
    {
        _f = value;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(double)
     *  \param[in]  value   浮動小数点数
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(double value) noexcept
        : _type(Type::Float)
        , _ui(0)
        , _string()
    {
        _f = static_cast<float>(value);
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(bool)
     *  \param[in]  value   論理値
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(bool value) noexcept
        : _type(Type::Bool)
        , _ui(0)
        , _string()
    {
        _b = value;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(STL::string)
     *  \param[in]  value   文字列
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(const STL::string &value) noexcept
        : _type(Type::String)
        , _i(0)
        , _string(value)
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(const char *)
     *  \param[in]  value   文字列
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(const char *value) noexcept
        : _type(Type::String)
        , _i(0)
        , _string(value)
    {}

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ(const void *)
     *  \param[in]  value   ポインタ変数
     *  \note
     *      アドレスを出力する．
     *      const char *型は文字列として認識されるため，void *型にキャストする必要がある．
     */
    /* ------------------------------------------------------------------------- */
    FormatArgument(const void *value) noexcept
        : _type(Type::Pointer)
        , _ui(reinterpret_cast<uint64_t>(value))
        , _string()
    {}

    //! 変換オプション
    struct Option
    {
        //! 値の表示モード
        enum class ValueDisplayMode
        {
            Decimal,            //!< 10進数（デフォルト）
            UpperCaseHex,       //!< 大文字の16進数
            LowerCaseHex        //!< 小文字の16進数
        };

        bool isValid;           //!< オプションの有効フラグ
        int index;              //!< インデックス
        int width;              //!< 表示幅（負数で左寄せ）
        int precision;          //!< 小数点以下の表示桁数
        bool showSeparator;     //!< 区切り文字を表示
        bool showPlusSign;      //!< 正の数にも符号を表示
        char fillCharacter;     //!< 表示幅を指定した際に埋める文字の指定
        bool skipAutoIndex;     //!< インデックスの自動カウントをスキップ
        ValueDisplayMode valueDisplayMode;  //!< 値の表示モード
        int hexWidth;           //!< 16進数の表示桁数

        //! コンストラクタ
        constexpr Option() noexcept
            : isValid(false)
            , index(-1)
            , width(0)
            , precision(-1)
            , showSeparator(false)
            , showPlusSign(false)
            , fillCharacter(' ')
            , skipAutoIndex(false)
            , valueDisplayMode(ValueDisplayMode::Decimal)
            , hexWidth(0)
        {}
    };

    STL::string ToString(const Option &option = Option()) const noexcept;

    static const char *Parse(Option &option, const char *parseText) noexcept;
    
    static STL::vector<IndexedCharacter> ToIndexedCharacters(const Option &option) noexcept;
    
private:
    //! 値の型
    enum class Type
    {
        Void,           //!< 型なし
        Int,            //!< 符号付き整数
        UnsignedInt,    //!< 符号なし整数
        Float,          //!< 浮動小数点数
        Bool,           //!< 論理値
        String,         //!< 文字列
        Pointer,        //!< ポインタ
    };
    
    bool IsZero() const noexcept;
    bool IsPositiveValue() const noexcept;
    bool IsNegativeValue() const noexcept;
    STL::string ToStringFromNumber(const Option &option) const noexcept;
    STL::string ToStringFromDecimal(const Option &option) const noexcept;
    STL::string ToHexStringFromValue(const Option &option) const noexcept;

    STL::string AlignmentString(const Option &option, const STL::string &string) const noexcept;
    
    Type _type;
    union
    {
        int64_t _i;
        uint64_t _ui;
        float _f;
        bool _b;
    };

    STL::string _string;
};

//! テキストフォーマットの引数の配列
using FormatArgs = STL::vector<FormatArgument>;

}   // namespace MGL::Text

#endif	// INCGUARD_MGL_TEXT_FORMAT_ARGUMENT_H_1622208320

// vim: et ts=4 sw=4 sts=4
