// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_ui_event.h
 *  \brief      MGL UIイベント
 *  \date       Since: July 7, 2021. 10:57:22 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_UI_EVENT_H_1625623042
#define INCGUARD_MGL_UI_EVENT_H_1625623042

#include <mgl/math/mgl_rectangle.h>
#include <mgl/input/touch/mgl_touch.h>

namespace MGL::UI
{
//! イベントタイプ
enum class EventType
{
    None,       //!< なし
    Selected,   //!< 選択
    Unselect,   //!< 選択解除
    Decide,     //!< 決定
};

class Event;

//! イベントリスナーインターフェース
class EventListener
{
    friend Event;
    
protected:
    virtual ~EventListener() noexcept = default;
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      イベント受信時のコールバック処理
     *  \param[in]  identifier  イベント識別子
     *  \param[in]  eventType   イベントの種類
     */
    /* ------------------------------------------------------------------------- */
    virtual void OnReceiveUIEvent(int32_t identifier, EventType eventType) noexcept = 0;
};

//! イベントクラス
class Event
{
public:
    //! ステート
    enum class State
    {
        Unselect,   //!< 非選択
        Selected,   //!< 選択中
    };
    
    struct Work
    {
        State state;
        int32_t identifier;
        MGL::Input::TouchID touchID;
        
        constexpr Work() noexcept
            : state(State::Unselect)
            , identifier(-1)
            , touchID(MGL::Input::TouchID::Invalid)
        {
        }
    };
    
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     */
    /* ------------------------------------------------------------------------- */
    constexpr Event(EventListener *listener) noexcept
        : _listener(listener)
        , _available(true)
    {
    }

    virtual ~Event() noexcept = default;

    EventType Update(Work &work, const MGL::Rectangle &rectangle, bool requestCancel) noexcept;

    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      イベントの有効状態を設定
     *  \param[in]  available   有効状態のフラグ
     */
    /* ------------------------------------------------------------------------- */
    constexpr void SetAvailable(bool available) noexcept
    {
        _available = available;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      イベントの有効状態を取得
     *  \return     有効状態
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool IsAvailable() const noexcept
    {
        return _available;
    }
    
private:
    EventType UpdateMouseOperation(Work &work, const MGL::Rectangle &rectangle) noexcept;
    EventType UpdateTouchOperation(Work &work, const MGL::Rectangle &rectangle) noexcept;

    EventListener *_listener;
    bool _available;
};

}   // namespace MGL::UI

#endif	// INCGUARD_MGL)UI_EVENT_H_1625623042

// vim: et ts=4 sw=4 sts=4
