// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_ui_widget.h
 *  \brief      MGL UIウィジット
 *  \date       Since: July 5, 2021. 7:59:33 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MGL_UI_WIDGET_H_1625439573
#define INCGUARD_MGL_UI_WIDGET_H_1625439573

#include <mgl/ui/mgl_ui_event.h>
#include <mgl/common/mgl_alignment.h>
#include <mgl/mgl_environment.h>
#include <mgl/stl/mgl_stl_containers.h>

namespace MGL::UI
{
//! UIウィジットクラス
class Widget
{
public:
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     */
    /* ------------------------------------------------------------------------- */
    Widget(uint32_t typeIdentifier) noexcept
        : _typeIdentifier(typeIdentifier)
        , _parent(nullptr)
        , _children()
        , _isVisible(true)
        , _transparency(1.0f)
        , _event(nullptr)
        , _eventWork()
        , _isEnabledEvent(true)
        , _pivotAlignment()
        , _anchorAlignment()
    {
    }

    virtual ~Widget() noexcept = default;

    void Update(const MGL::Vector2 &offset = MGL::Vector2(0.0f, 0.0f), bool updateEvent = true) noexcept;
    void Render(const MGL::Vector2 &offset = MGL::Vector2(0.0f, 0.0f)) noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      更新処理（オフセットなし）
     *  \param[in]  updateEvent イベント更新を行うかのフラグ
     */
    /* ------------------------------------------------------------------------- */
    MGL_MAYBE_CONSTEXPR void Update(bool updateEvent) noexcept
    {
        return Update(MGL::Vector2(), updateEvent);
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      矩形の取得
     *  \return     このウィジットの矩形
     */
    /* ------------------------------------------------------------------------- */
    virtual MGL::Rectangle GetRectangle() const noexcept = 0;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ウィジットの種類の識別子を取得
     *  \return     識別子
     */
    /* ------------------------------------------------------------------------- */
    constexpr uint32_t GetTypeIdentifier() const noexcept
    {
        return _typeIdentifier;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      可視状態の設定
     *  \param[in]  isVisible   設定する可視状態．trueで可視
     */
    /* ------------------------------------------------------------------------- */
    constexpr void SetVisible(bool isVisible) noexcept
    {
        _isVisible = isVisible;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      このウィジットに設定された可視状態のフラグを取得
     *  \return     可視状態のフラグ
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool GetLocalVisibleFlag() const noexcept
    {
        return _isVisible;
    }

    bool IsVisible() const noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      透過値の設定
     *  \param[in]  transparency   設定する透過値
     */
    /* ------------------------------------------------------------------------- */
    constexpr void SetTransparency(float transparency) noexcept
    {
        _transparency = transparency;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      このウィジットに設定された透過値の設定
     *  \return     このウィジットの透過値
     */
    /* ------------------------------------------------------------------------- */
    constexpr float GetLocalTransparency() const noexcept
    {
        return _transparency;
    }

    float GetTransparency() const noexcept;

    MGL::Vector2 GetPosition() const noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      サイズを取得
     *  \return     サイズ
     */
    /* ------------------------------------------------------------------------- */
    constexpr MGL::Vector2 GetSize() const noexcept
    {
        return GetRectangle().GetSize();
    }

    MGL::Vector2 GetParentSize() const noexcept;
    
    bool AddChild(Widget *child) noexcept;
    bool RemoveChild(Widget *child) noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      親ウィジットの取得
     *  \return     親ウィジット
     */
    /* ------------------------------------------------------------------------- */
    const Widget *GetParent() const noexcept
    {
        return _parent;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      子ウィジットの取得
     *  \return     子ウィジットのリスト
     */
    /* ------------------------------------------------------------------------- */
    const STL::list<Widget *> &GetChildren() const noexcept
    {
        return _children;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      イベントを設定
     *  \param[in]  event   設定するイベント
     */
    /* ------------------------------------------------------------------------- */
    constexpr void SetEvent(Event *event) noexcept
    {
        _event = event;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      イベントを取得
     *  \return     イベント
     */
    /* ------------------------------------------------------------------------- */
    constexpr Event *GetEvent() const noexcept
    {
        return _event;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      イベントIDを設定
     *  \param[in]  identifier  設定するイベントID
     */
    /* ------------------------------------------------------------------------- */
    constexpr void SetEventIdentifier(int32_t identifier) noexcept
    {
        _eventWork.identifier = identifier;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      イベントIDを取得
     *  \return     イベントID
     */
    /* ------------------------------------------------------------------------- */
    constexpr int32_t GetEventIdentifier() const noexcept
    {
        return _eventWork.identifier;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      イベントを有効にするかを設定
     *  \param[in]  isEnabled   trueで有効設定
     */
    /* ------------------------------------------------------------------------- */
    constexpr void SetEnabledEvent(bool isEnabled) noexcept
    {
        _isEnabledEvent = isEnabled;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      イベントが有効状態であるかを取得
     *  \retval     true    有効
     *  \retval     false   無効
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool IsEnabledEvent() const noexcept
    {
        return _isEnabledEvent;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ウィジットの中心アライメントを設定
     *  \param[in]  pivotAlignment  中心位置のアライメント
     */
    /* ------------------------------------------------------------------------- */
    constexpr void SetPivot(const MGL::Alignment &pivotAlignment) noexcept
    {
        _pivotAlignment = pivotAlignment;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ウィジットの中心アライメントを取得
     *  \return     中心位置のアライメント
     */
    /* ------------------------------------------------------------------------- */
    constexpr const MGL::Alignment &GetPivot() const noexcept
    {
        return _pivotAlignment;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ウィジットのアンカーを設定
     *  \param[in]  anchorAlignment     アンカー設定
     */
    /* ------------------------------------------------------------------------- */
    constexpr void SetAnchor(const MGL::Alignment &anchorAlignment) noexcept
    {
        _anchorAlignment = anchorAlignment;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ウィジットのアンカー取得
     *  \return     アンカー設定
     */
    /* ------------------------------------------------------------------------- */
    constexpr const MGL::Alignment &GetAnchor() const noexcept
    {
        return _anchorAlignment;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      イベントのステートを取得
     *  \return     イベントのステート
     */
    /* ------------------------------------------------------------------------- */
    constexpr Event::State GetEventState() const noexcept
    {
        return _eventWork.state;
    }
    
protected:
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      更新処理
     *  \param[in]  position    このウィジットの位置
     */
    /* ------------------------------------------------------------------------- */
    virtual void OnUpdate(const MGL::Vector2 &position) noexcept
    {
        (void)position;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      描画処理
     *  \param[in]  position    このウィジットの位置
     */
    /* ------------------------------------------------------------------------- */
    virtual void OnRender(const MGL::Vector2 &position) noexcept
    {
        (void)position;
    }
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      イベント通知後の処理
     *  \param[in]  eventType   通知したイベントの種類
     */
    /* ------------------------------------------------------------------------- */
    virtual void OnNotifiedEvent(EventType eventType) noexcept
    {
        (void)eventType;
    }

private:
    void UpdateEvent(const MGL::Vector2 &offset, bool &cancelRequest) noexcept;
        
    MGL::Vector2 GetAdjustedPosition() const noexcept;

    uint32_t _typeIdentifier;
    
    Widget *_parent;
    STL::list<Widget *> _children;

    bool _isVisible;
    float _transparency;
    
    Event *_event;
    Event::Work _eventWork;
    bool _isEnabledEvent;

    MGL::Alignment _pivotAlignment;
    MGL::Alignment _anchorAlignment;
};

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      型情報付きのウィジットクラス
 *  \tparam     Type    識別子に関連付ける型
 */
/* ------------------------------------------------------------------------- */
template <typename Type>
class TypedWidget : public Widget
{
public:
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     *  \param[in]  type    識別子
     */
    /* ------------------------------------------------------------------------- */
    TypedWidget(Type type) noexcept
        : Widget(static_cast<uint32_t>(type))
    {
    }
    
    virtual ~TypedWidget() noexcept = default;
    
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      識別子を取得
     *  \return     識別子
     */
    /* ------------------------------------------------------------------------- */
    constexpr Type GetType() const noexcept
    {
        return static_cast<Type>(GetTypeIdentifier());
    }
};
}   // namespace MGL::UI

#endif	// INCGUARD_MGL_UI_WIDGET_H_1625439573

// vim: et ts=4 sw=4 sts=4
