// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_singleton.cc
 *  \brief      MGL シングルトンクラス
 *  \date       Since: July 17, 2022. 4:09:40 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/common/mgl_singleton.h>

namespace MGL
{
namespace
{
static inline std::list<SingletonFinalizer::FinalizeFunction> sFunctions;   // staticなのでMGL::STL::listは使用できない
static inline std::recursive_mutex sMutex;
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      シングルトンクラスの登録
 *  \param[in]  function    解放処理の関数
 */
/* ------------------------------------------------------------------------- */
void SingletonFinalizer::Register(FinalizeFunction function) noexcept
{
    std::lock_guard lock(sMutex);

    sFunctions.push_back(function);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      登録されたシングルトンクラスの解放
 */
/* ------------------------------------------------------------------------- */
void SingletonFinalizer::Finalize() noexcept
{
    std::lock_guard lock(sMutex);
    
    while (!sFunctions.empty())
    {
        sFunctions.back()();
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      シングルトンクラスの登録解除
 *  \param[in]  function    解放処理の関数
 */
/* ------------------------------------------------------------------------- */
void SingletonFinalizer::Unregister(FinalizeFunction function) noexcept
{
    std::lock_guard lock(sMutex);
    
    auto it = std::find(sFunctions.begin(), sFunctions.end(), function);
    if (it != sFunctions.end())
    {
        sFunctions.erase(it);
    }
}
}   // namespace MGL

// vim: et ts=4 sw=4 sts=4
