// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_file_delegate_cocoa.mm
 *  \brief      MGL iOS/tvOS用ファイルデリゲート
 *  \date       Since: February 27, 2021. 19:08:45 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/file/delegate/mgl_file_delegate_cocoa.h>
#if defined(MGL_FILE_DELEGATE_ENABLE_POSIX) && defined(MGL_FILE_DELEGATE_ENABLE_COCOA)

#include <Foundation/Foundation.h>

namespace MGL::File
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ディレクトリを作成する
 *  \param[in,out]  mountWork   マウント毎のワーク
 *  \param[in]      path        作成するディレクトリのパス
 *  \return         エラー発生時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
Result CocoaDelegate::MakeDirectory([[maybe_unused]] MountWork *mountWork, const char *path) noexcept
{
    NSString *dirPath = [NSString stringWithUTF8String:path];
    
    auto *manager = NSFileManager.defaultManager;

    // 既に存在しているかをチェック
    BOOL isDirectory = NO;
    auto existsFile = [manager fileExistsAtPath:dirPath
                               isDirectory:&isDirectory];
    if (existsFile)
    {
        // ディレクトリが既に存在しているのであれば成功
        if (isDirectory)
        {
            return Result::Succeeded();
        }
        // ディレクトリでないファイルが存在していたら失敗
        else
        {
            return Error::FileAlreadyExist;
        }
    }
    
    // ディレクトリを生成
    NSError *error = nil;
    auto isSucceeded = [manager createDirectoryAtPath:dirPath
                          withIntermediateDirectories:NO
                                           attributes:nil
                                                error:&error];
    // 失敗していたらエラーコードを取得して返す
    if (!isSucceeded)
    {
        if (error.domain == NSPOSIXErrorDomain)
        {
            return GetError(static_cast<int>(error.code));
        }
        
        return Error::UnknownError;
    }
    
    return Result::Succeeded();
}
}   // namespace MGL::File

#endif  // defined(MGL_FILE_DELEGATE_ENABLE_POSIX) && defined(MGL_FILE_DELEGATE_ENABLE_COCOA)
// vim: et ts=4 sw=4 sts=4
