// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_file_delegate_posix.cc
 *  \brief      MGL ファイルデリゲート POSIX準拠
 *  \date       Since: January 20, 2021. 11:37:34 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/file/delegate/mgl_file_delegate_posix.h>
#if defined(MGL_FILE_DELEGATE_ENABLE_POSIX)

#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>

namespace MGL::File
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ファイルをオープン
 *  \param[out]     result      処理結果
 *  \param[in,out]  mountWork   マウント毎のワーク
 *  \param[in]      path        オープンするファイルのパス
 *  \param[in]      mode        オープンモード
 *  \return         オープンしたファイルのワーク
 */
/* ------------------------------------------------------------------------- */
AccessWorkPtr POSIXDelegate::Open(Result &result, [[maybe_unused]] SharedMountWork &mountWork, const char *path, OpenMode mode) noexcept
{
    struct stat st;

    // 読み込み時のチェック
    if (mode == OpenMode::Read)
    {
        // ファイルの状態を取得
        if (stat(path, &st) != 0)
        {
            result = GetError(errno);
            return nullptr;
        }
        
        // 通常ファイルでなければ失敗
        if (!S_ISREG(st.st_mode))
        {
            result = Error::NoRegularFile;
            return nullptr;
        }
    }
    // 書き込み時のチェック
    else if (mode == OpenMode::Write)
    {
        // ファイルの状態をチェック
        if (stat(path, &st) != 0)
        {
            // ファイルが存在しない場合はエラーにしない
            result = GetError(errno);
            if (result != Error::FileNotExist)
            {
                return nullptr;
            }
        }
        else
        {
            // 通常ファイルでなければ失敗
            if (!S_ISREG(st.st_mode))
            {
                result = Error::NoRegularFile;
                return nullptr;
            }
        }
        
    }
    // それ以外
    else
    {
        // 引数が不正
        result = Error::InvalidArgument;
        return nullptr;
    }
    
    // ファイルをオープン
    FILE *fp = fopen(path, GetModeString(mode));
    if (fp == nullptr)
    {
        result = GetError(errno);
        return nullptr;
    }

    // ワークに値を設定
    auto work = STL::make_unique<POSIXWork>(mode);
    work->fp = fp;
    work->size = static_cast<size_t>(st.st_size);
    
    result.Success();

    return work;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ファイルをクローズ
 *  \param[in,out]  work    クローズするファイルのワーク
 *  \return         エラー発生時にいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
Result POSIXDelegate::Close(AccessWork *work) noexcept
{
    auto *posixWork = GetWork(work);
    if (posixWork == nullptr)
    {
        return Error::InvalidArgument;
    }
    
    fclose(posixWork->fp);
    posixWork->fp = nullptr;
    
    return Result::Succeeded();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ファイルを読み込み
 *  \param[in,out]  work        アクセスするファイルのワーク
 *  \param[out]     result      処理結果
 *  \param[in]      buffer      読み込み先のアドレス
 *  \param[in]      size        読み込みサイズ
 *  \return         実際に読み込んだバイト数
 */
/* ------------------------------------------------------------------------- */
size_t POSIXDelegate::Read(AccessWork *work, Result &result, void *buffer, size_t size) noexcept
{
    auto *posixWork = GetWork(work, OpenMode::Read);
    if (posixWork == nullptr)
    {
        result = Error::InvalidArgument;
        return 0;
    }

    clearerr(posixWork->fp);
    auto readSize = fread(buffer, 1, size, posixWork->fp);

    if (ferror(posixWork->fp) != 0)
    {
        result = Error::ReadError;
    }

    result.Success();
    return readSize;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ファイルを書き込み
 *  \param[in,out]  work        アクセスするファイルのワーク
 *  \param[out]     result      処理結果
 *  \param[out]     buffer      書き込み先のアドレス
 *  \param[in]      size        書き込みサイズ
 *  \return         実際に書き込んだバイト数
 */
/* ------------------------------------------------------------------------- */
size_t POSIXDelegate::Write(AccessWork *work, Result &result, const void *buffer, size_t size) noexcept
{
    auto *posixWork = GetWork(work, OpenMode::Write);
    if (posixWork == nullptr)
    {
        result = Error::InvalidArgument;
        return 0;
    }

    clearerr(posixWork->fp);
    auto writeSize = fwrite(buffer, 1, size, posixWork->fp);
    if (writeSize < size)
    {
        result = Error::WriteError;
    }
    else
    {
        result.Success();
    }
    
    return writeSize;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ストリーム位置を設定
 *  \param[in,out]  work        アクセスするファイルのワーク
 *  \param[out]     result      エラー発生時の結果の格納先
 *  \param[in]      seekType    シークタイプ
 *  \param[in]      offset      オフセット
 *  \return         エラー発生時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
size_t POSIXDelegate::Seek(AccessWork *work, Result &result, SeekType seekType, int32_t offset) noexcept
{
    auto *posixWork = GetWork(work);
    if (posixWork == nullptr)
    {
        result = Error::InvalidArgument;
        return 0;
    }

    int seek = SEEK_SET;    
    switch (seekType)
    {
        case SeekType::Top:
            seek = SEEK_SET;
            break;
            
        case SeekType::End:
            seek = SEEK_END;
            break;
            
        case SeekType::Current:
            seek = SEEK_CUR;
            break;
    }
    
    if (fseek(posixWork->fp, offset, seek) != 0)
    {
        result = GetError(errno);
    }
    else
    {
        result.Success();
    }

    return static_cast<size_t>(ftell(posixWork->fp));
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ストリーム位置を取得
 *  \param[in]  work    アクセスするファイルのワーク
 *  \param[out] result  エラー発生時の結果の格納先
 *  \return     現在のストリーム位置
 */
/* ------------------------------------------------------------------------- */
size_t POSIXDelegate::GetOffset(AccessWork *work, Result &result) const noexcept
{
    auto *posixWork = GetWork(work);
    if (posixWork == nullptr)
    {
        result = Error::InvalidArgument;
        return 0;
    }
    
    auto offset = ftell(posixWork->fp);
    if (offset < 0)
    {
        result = Error::UnknownError;
        return 0;
    }

    result.Success();
    return static_cast<size_t>(offset);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ファイルストリームが終端に達しているかを取得
 *  \param[in,out]  work    アクセスするファイルのワーク
 *  \param[out]     result  エラー発生時の結果の格納先
 *  \retval         true    終端に達している
 *  \retval         false   終端に達していない
 */
/* ------------------------------------------------------------------------- */
bool POSIXDelegate::IsEOF(AccessWork *work, Result &result) const noexcept
{
    auto *posixWork = GetWork(work);
    if (posixWork == nullptr)
    {
        result = Error::InvalidArgument;
        return false;
    }

    auto offset = GetOffset(work, result);
    if (result.HasError())
    {
        return false;
    }
    if (offset >= posixWork->size)
    {
        return true;
    }
    
    return false;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      オープンしているファイルのサイズを取得
 *  \param[in]  work    アクセスするファイルのワーク
 *  \param[out] result  エラー発生時の結果の格納先
 *  \return     ファイルサイズ
 */
/* ------------------------------------------------------------------------- */
size_t POSIXDelegate::GetSize(AccessWork *work, Result &result) const noexcept
{
    auto *posixWork = GetWork(work);
    if (posixWork == nullptr)
    {
        result = Error::InvalidArgument;
        return 0;
    }

    auto offset = ftell(posixWork->fp);
    fseek(posixWork->fp, 0, SEEK_END);
    auto size = ftell(posixWork->fp);
    fseek(posixWork->fp, offset, SEEK_SET);
    
    if (size < 0)
    {
        result = Error::UnknownError;
        return 0;
    }

    result.Success();
    return static_cast<size_t>(size);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ファイルのサイズを取得
 *  \param[in,out]  mountWork   マウント毎のワーク
 *  \param[out]     result      エラー発生時の結果の格納先
 *  \param[in]      path        サイズを取得するファイルのパス
 *  \return         ファイルサイズ
 */
/* ------------------------------------------------------------------------- */
size_t POSIXDelegate::GetSize([[maybe_unused]] MountWork *mountWork, Result &result, const char *path) const noexcept
{
    struct stat st;
    
    if (stat(path, &st) != 0)
    {
        result = GetError(errno);
        return 0;
    }
    
    result.Success();
    return static_cast<size_t>(st.st_size);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ディレクトリを作成する
 *  \param[in,out]  mountWork   マウント毎のワーク
 *  \param[in]      path        作成するディレクトリのパス
 *  \return         エラー発生時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
Result POSIXDelegate::MakeDirectory([[maybe_unused]] MountWork *mountWork, const char *path) noexcept
{
    struct stat st;
    
    // 既存のファイルをチェック
    if (stat(path, &st) == 0)
    {
        // 既存のファイルがディレクトリであれば成功とする
        if (S_ISDIR(st.st_mode))
        {
            return Result::Succeeded();
        }

        return Error::FileAlreadyExist;
    }

    // ディレクトリを生成
    if (mkdir(path, S_IRWXU | S_IRGRP | S_IXGRP) == -1)
    {
        return GetError(errno);
    }

    return Result::Succeeded();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ファイルの移動・リネーム
 *  \param[in,out]  mountWork   マウント毎のワーク
 *  \param[in]      sourcePath  移動元のパス
 *  \param[in]      destPath    移動先のパス
 *  \return         エラー発生時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
Result POSIXDelegate::Move([[maybe_unused]] MountWork *mountWork, const char *sourcePath, const char *destPath) noexcept
{
    int result = rename(sourcePath, destPath);
    if (result == -1)
    {
        return GetError(errno);
    }

    return Result::Succeeded();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ファイルの削除
 *  \param[in,out]  mountWork   マウント毎のワーク
 *  \param[in]      path        削除するファイルのパス
 *  \return         エラー発生時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
Result POSIXDelegate::Remove([[maybe_unused]] MountWork *mountWork, const char *path) noexcept
{
    int result = remove(path);
    if (result != 0)
    {
        return GetError(errno);
    }

    return Result::Succeeded();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ファイルの存在をチェック
 *  \param[in,out]  mountWork   マウント毎のワーク
 *  \param[out]     result      エラー発生時の結果の格納先
 *  \param[in]      path        チェックするファイルのパス
 *  \return         存在する場合はエラーなしのリザルトが返る
 */
/* ------------------------------------------------------------------------- */
bool POSIXDelegate::Exists([[maybe_unused]] MountWork *mountWork, Result &result, const char *path) noexcept
{
    struct stat st;
    if (stat(path, &st) == 0)
    {
        result.Success();
        return true;
    }
    
    if (errno == ENOENT)
    {
        result.Success();
    }
    else
    {
        result = GetError(errno);
    }
    
    return false;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      作業用構造体からこのクラス用の構造体を取得する
 *  \param[in]  work    作業用構造体
 *  \param[in]  mode    オープンモード指定．OpenMode::Noneを指定した場合は無視される
 *  \return     作業用構造体がこのクラス用に作られていれば有効なアドレスを返す
 */
/* ------------------------------------------------------------------------- */
POSIXDelegate::POSIXWork *POSIXDelegate::GetWork(AccessWork *work, OpenMode mode) const noexcept
{
    if (work != nullptr)
    {
        if (work->IsOpen())
        {
            if ((mode == OpenMode::None) || (mode == work->GetOpenMode()))
            {
                if (work->GetDelegateKey() == kDelegateKey)
                {
                    return static_cast<POSIXWork *>(work);
                }
            }
        }
    }
    
    return nullptr;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      オープンモードからモード指定用の文字列を取得
 *  \param[in]  mode    オープンモード
 *  \return     オープンモードに対応した文字列
 */
/* ------------------------------------------------------------------------- */
const char *POSIXDelegate::GetModeString(OpenMode mode) const noexcept
{
    static const char *modeReadBinary = "rb";
    static const char *modeWriteBinary = "wb";

    switch (mode)
    {
        case OpenMode::Read:
            return modeReadBinary;

        case OpenMode::Write:
            return modeWriteBinary;

        default:
            break;
    }

    return nullptr;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      エラー番号から処理結果を取得する
 *  \param[in]  errorNumber     ファイル操作後のerrno
 *  \return     エラー番号に対応した処理結果
 */
/* ------------------------------------------------------------------------- */
Error POSIXDelegate::GetError(int errorNumber) const noexcept
{
    switch (errorNumber)
    {
        case ENOENT:
            return Error::FileNotExist;
            
        case EPERM:
        case EACCES:
            return Error::NotPermitted;
            
        case EINVAL:
            return Error::InvalidArgument;
            
        default:
            break;
    }
    
    return Error::UnknownError;
}

}   // namespace MGL::File
#endif  // MGL_FILE_DELEGATE_ENABLE_POSIX

// vim: et ts=4 sw=4 sts=4
