// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_file_utility.cc
 *  \brief      MGL ファイルユーティリティ
 *  \date       Since: September 29, 2022. 15:40:31 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/file/mgl_file_utility.h>

#include <mgl/file/mgl_file_accessor.h>

namespace MGL::File
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ファイルのサイズを取得
 *  \param[in]  path    サイズを取得するファイルのパス
 *  \return     ファイルサイズ
 */
/* ------------------------------------------------------------------------- */
size_t Utility::GetSize(const char *path) noexcept
{
    return Accessor::GetSize(_result, path);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ディレクトリを作成する
 *  \param[in]  path    作成するディレクトリのパス
 *  \return     失敗時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
Result Utility::MakeDirectory(const char *path) const noexcept
{
    return Accessor::MakeDirectory(path);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ファイルの移動・リネーム
 *  \param[in]  sourcePath  移動元のパス
 *  \param[in]  destPath    移動先のパス
 *  \return     失敗時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
Result Utility::Move(const char *sourcePath, const char *destPath) const noexcept
{
    return Accessor::Move(sourcePath, destPath);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ファイルの削除
 *  \param[in]  path    削除するファイルのパス
 *  \return     失敗時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
Result Utility::Remove(const char *path) const noexcept
{
    return Accessor::Remove(path);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ファイルのコピー
 *  \param[in]  sourcePath  コピー元のパス
 *  \param[in]  destPath    コピー先のパス
 *  \return     失敗時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
Result Utility::Copy(const char *sourcePath, const char *destPath) const noexcept
{
    return Accessor::Copy(sourcePath, destPath);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ファイルの存在をチェック
 *  \param[in]  path        チェックするファイルのパス
 *  \retval     true        存在する
 *  \retval     false       存在しない
 *  \note       失敗時には戻り値にいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
bool Utility::Exists(const char *path) noexcept
{
    return Accessor::Exists(_result, path);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      システム標準のファイルであるかをチェック
 *  \param[in]  path    パス
 *  \retval     true    システム標準のファイルである
 *  \retval     false   システム標準のファイルではない
 *  \note       失敗時には戻り値にいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
bool Utility::IsSystemNativeFile(const char *path) noexcept
{
    return Accessor::IsSystemNativeFile(_result, path);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      マウント
 *  \param[in]  mountName   マウント名
 *  \param[in]  path        マウント先のパス
 *  \param[in]  accessType  アクセスタイプ
 *  \param[in]  delegateKey デリゲートキー
 *  \return     失敗時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
Result Utility::Mount(const char *mountName, const char *path, MountAccessType accessType, DelegateKey delegateKey) const noexcept
{
    return Accessor::Mount(mountName, path, accessType, delegateKey);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      マウント解除
 *  \param[in]  mountName   マウント名
 *  \return     失敗時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
Result Utility::Unmount(const char *mountName) const noexcept
{
    return Accessor::Unmount(mountName);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      再マウント
 *  \param[in]  mountName   マウント名
 *  \param[in]  path        マウント先のパス
 *  \param[in]  accessType  アクセスタイプ
 *  \param[in]  delegateKey デリゲートキー
 *  \return     失敗時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
Result Utility::Remount(const char *mountName, const char *path, MountAccessType accessType, DelegateKey delegateKey) const noexcept
{
    return Accessor::Remount(mountName, path, accessType, delegateKey);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      マウントされているかを取得
 *  \param[in]  mountName   マウント名
 *  \retval     true        マウントされている
 *  \retval     false       マウントされていない
 */
/* ------------------------------------------------------------------------- */
bool Utility::IsMounted(const char *mountName) const noexcept
{
    return Accessor::IsMounted(mountName);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      マウントパスからシステム標準のパスに変換
 *  \param[in]  path                パス
 *  \return     システム標準のパス．失敗時にはいずれかのエラーが設定される
 */
/* ------------------------------------------------------------------------- */
STL::string Utility::GetSystemNativePath(const char *path) noexcept
{
    return Accessor::GetSystemNativePath(_result, path);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      デリゲートの削除
 *  \param[in]  key                 削除するデリゲートキー
 *  \exception  MGL::File::Result   失敗時にエラーが設定されたリザルトを返す
 *  \note       存在しないデリゲートキーに対しては何もしない
 */
/* ------------------------------------------------------------------------- */
Result Utility::RemoveDelegate(DelegateKey key) const noexcept
{
    return Accessor::RemoveDelegate(key);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      デフォルトのデリゲートを設定
 *  \param[in]  key     デフォルトのデリゲートキー
 *  \note       この関数は例外を発生させず，File::SetDefaultDelegate()と等価である
 */
/* ------------------------------------------------------------------------- */
void Utility::SetDefaultDelegate(DelegateKey key) const noexcept
{
    Accessor::SetDefaultDelegate(key);
}
}   // namespace MGL::File

// vim: et ts=4 sw=4 sts=4
