// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_initializer_ios.mm
 *  \brief      MGL iOS用イニシャライザクラス
 *  \date       Since: March 19, 2021. 18:09:41 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/initialize/mgl_initializer_ios.h>
#if defined(MGL_TARGET_IOS)

#include <Foundation/Foundation.h>

#include <mgl/audio/renderer/mgl_audio_renderer_apple_coreaudio.h>

#include <mgl/file/delegate/mgl_file_delegate_cocoa.h>
#include <mgl/file/mgl_file.h>

#include <mgl/input/gamepad/mgl_gamepad_delegate_apple_gc.h>
#include <mgl/input/gamepad/mgl_gamepad_server.h>
#include <mgl/input/touch/mgl_touch_delegate_event.h>
#include <mgl/input/keyboard/mgl_keyboard_delegate_apple_gc.h>
#include <mgl/input/mouse/mgl_mouse_delegate_apple_gc.h>

#include <mgl/render/metal/mgl_metal_defs.h>
#include <mgl/render/metal/mgl_metal_renderer_2d.h>
#include <mgl/render/metal/mgl_metal_texture_generator.h>
#include <mgl/render/metal/mgl_metalkit_texture_loader.h>
#include <mgl/render/mgl_render.h>

#include <mgl/system/application/mgl_application_module_uikit.h>
#include <mgl/system/chrono/mgl_chrono_module_apple.h>
#include <mgl/system/debug/mgl_debug_module_apple.h>
#include <mgl/system/locale/mgl_locale_module_apple.h>

#include <mgl/system/mgl_system_debug_macro.h>

namespace MGL::Apple::iOS
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ
 */
/* ------------------------------------------------------------------------- */
Initializer::Initializer() noexcept
    : _isEnabledGCGamepadDelegate(true)
    , _isEnabledKeyboard(true)
{
    SetAudioInitializeMode(Audio::InitializeMode::Sample44k2ch);
    SetLogLevel(System::LogLevel::AppTrace);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      2Dレンダラを生成
 *  \return     2Dレンダラ
 */
/* ------------------------------------------------------------------------- */
STL::unique_ptr<Render::Renderer2DDelegate> Initializer::MakeRenderer2D() const noexcept
{
    return STL::make_unique<Render::Metal::Renderer2D>();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      テクスチャジェネレータを生成
 *  \return     テクスチャジェネレータ
 */
/* ------------------------------------------------------------------------- */
STL::unique_ptr<Render::TextureGenerator> Initializer::MakeTextureGenerator() const noexcept
{
    return STL::make_unique<Render::Metal::TextureGenerator>();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      テクスチャローダを生成
 */
/* ------------------------------------------------------------------------- */
void Initializer::MakeTextureLoader(Render::TextureStorage &textureStorage) const noexcept
{
    // MetalKitによるテクスチャローダーを追加
    textureStorage.RegisterLoader(Render::Metal::MetalKitTextureLoader::kLoaderKey, STL::make_unique<Render::Metal::MetalKitTextureLoader>());
    textureStorage.SetDefaultLoader(Render::Metal::MetalKitTextureLoader::kLoaderKey);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      オーディオレンダラの生成
 *  \return     オーディオレンダラ
 */
/* ------------------------------------------------------------------------- */
STL::unique_ptr<Audio::Renderer> Initializer::MakeAudioRenderer() const noexcept
{
    return STL::make_unique<Audio::AppleCoreAudioRenderer>();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ファイルシステム初期化後の処理
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
File::Result Initializer::DidInitializeFileSystem() noexcept
{
    try
    {
        File::ThrowingUtility utility;
        
        // iOS/tvOS用デリゲートを追加してデフォルトに設定
        utility.AddDelegate<File::CocoaDelegate>(File::CocoaDelegate::kDelegateKey);
        utility.SetDefaultDelegate(File::CocoaDelegate::kDelegateKey);
        
        // リソースディレクトリをマウント
        utility.Mount("resource", NSBundle.mainBundle.resourcePath.UTF8String, File::MountAccessType::ReadOnly);
        
        // テンポラリディレクトリをマウント
        utility.Mount("temp", NSTemporaryDirectory().UTF8String, File::MountAccessType::Writable);
        
        // ユーザーディレクトリをマウント
        NSArray *documentPath = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
        if (documentPath.count == 0)
        {
            return File::Error::MountFailed;
        }
        utility.Mount("user", [documentPath.firstObject UTF8String], File::MountAccessType::Writable);
        
        return File::Result::Succeeded();
    }
    catch (File::Result result)
    {
        return result;
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      アプリケーションモジュールの生成
 *  \return     アプリケーションモジュール
 */
/* ------------------------------------------------------------------------- */
STL::unique_ptr<System::ApplicationModuleInterface> Initializer::MakeApplicationModule() const noexcept
{
    return STL::make_unique<System::UIKitApplicationModule>();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      日付と時刻のモジュールを生成
 *  \return     日付と時刻のモジュール
 */
/* ------------------------------------------------------------------------- */
STL::unique_ptr<System::ChronoModuleInterface> Initializer::MakeChronoModule() const noexcept
{
    return STL::make_unique<System::AppleChronoModule>();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      デバッグモジュールを生成
 *  \return     デバッグモジュール
 */
/* ------------------------------------------------------------------------- */
STL::unique_ptr<System::DebugModuleInterface> Initializer::MakeDebugModule() const noexcept
{
    return STL::make_unique<System::AppleDebugModule>(GetLogLevel());
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ロケール情報モジュールを生成
 *  \return     ロケール情報モジュール
 */
/* ------------------------------------------------------------------------- */
STL::unique_ptr<System::LocaleModuleInterface> Initializer::MakeLocaleModule() const noexcept
{
    return STL::make_unique<System::AppleLocaleModule>();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      キーボード入力デリゲートの生成
 *  \return     キーボード入力デリゲート
 */
/* ------------------------------------------------------------------------- */
STL::unique_ptr<Input::KeyboardDelegate> Initializer::MakeKeyboardDelegate() const noexcept
{
    if (_isEnabledKeyboard)
    {
        if (Input::AppleGCKeyboardDelegate::IsAvailable())
        {
            return STL::make_unique<Input::AppleGCKeyboardDelegate>();
        }
    }
    
    return nullptr;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      マウス入力デリゲートの生成
 *  \return     マウス入力デリゲート
 */
/* ------------------------------------------------------------------------- */
STL::unique_ptr<Input::MouseDelegate> Initializer::MakeMouseDelegate() const noexcept
{
#if 0
    if (Input::AppleGCMouseDelegate::IsAvailable())
    {
        return STL::make_unique<Input::AppleGCMouseDelegate>();
    }
#endif
    
    return nullptr;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      タッチ入力デリゲートの生成
 *  \return     タッチ入力デリゲート
 */
/* ------------------------------------------------------------------------- */
STL::unique_ptr<Input::TouchDelegate> Initializer::MakeTouchDelegate() const noexcept
{
    return STL::make_unique<Input::EventTouchDelegate>();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ゲームパッドサーバ初期化後の処理
 *  \param[in]  server      ゲームパッドサーバ
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool Initializer::DidInitializeGamepadServer(Input::GamepadServer &server) const noexcept
{
    // GCゲームパッド用デリゲートを追加
    if (_isEnabledGCGamepadDelegate)
    {
        if (!server.AddDelegate(STL::make_unique<Input::AppleGCGamepadDelegate>(server)))
        {
            MGL_WARNING("[MGL Gamepad] Failed to initialize GC gamepad delegate.");
        }
    }

    return true;
}


}   // namespace MGL::Apple::iOS

#endif  // MGL_TARGET_IOS
// vim: et ts=4 sw=4 sts=4
