// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_gamepad_state.cc
 *  \brief      MGL ゲームパッドステート
 *  \date       Since: January 7, 2021. 12:39:51 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/input/gamepad/mgl_gamepad_state.h>

#include <mgl/event/mgl_event.h>
#include <mgl/input/gamepad/mgl_gamepad_server.h>
#include <mgl/system/mgl_system_debug_macro.h>
#include <mgl/system/mgl_system_application.h>

namespace MGL::Input
{
namespace
{
constexpr float kAnalogStickThreshold = 0.3f;       // アナログスティックの閾値

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ユニークIDの生成
 *  \return     新たに生成されたユニークID
 */
/* ------------------------------------------------------------------------- */
PadID MakeUniqueID() noexcept
{
    static PadID sUniqueID = PadID::UniqueIDStart;

    auto newID = sUniqueID;

    sUniqueID = PadID{static_cast<std::underlying_type<PadID>::type>(sUniqueID) + 1};
    if (sUniqueID >= PadID::UniqueIDEnd)
    {
        sUniqueID = PadID::UniqueIDStart;
    }

    return newID;
}
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ
 */
/* ------------------------------------------------------------------------- */
PadState::PadState() noexcept
    : _type(PadType::Disable)
    , _identify(PadID::Invalid)
    , _deviceInfo()
    , _isUsing(false)
    , _buttonFlags(0)
    , _prevButtonFlags(0)
    , _repeatFlags(0)
    , _leftStick(0.0f, 0.0f)
    , _rightStick(0.0f, 0.0f)
    , _playerIndex(-1)
    , _entry(PadEntry::NoEntry)
    , _priority(PadPriority::Low)
    , _decideButton(PadButton::None)
    , _cancelButton(PadButton::None)
    , _isPortraitMode(false)
    , _repeatTimerArray()
{
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      更新前の処理
 */
/* ------------------------------------------------------------------------- */
void PadState::PreUpdate() noexcept
{
    if (IsEnabled())
    {
        // 要求があればアクティベート解除
        if (_shouldDeactivate)
        {
            Deactivate();
            _shouldDeactivate = false;
        }

        _prevButtonFlags = _buttonFlags;
        _buttonFlags = 0;
        _leftStick = _rightStick = Vector2(0.0f, 0.0f);
    }
    else
    {
        // 要求があればアクティベート
        if (_shouldActivate)
        {
            Activate();
            _shouldActivate = false;
            _buttonFlags = _prevButtonFlags = 0;
            _leftStick = _rightStick = Vector2(0.0f, 0.0f);
        }
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      アクティベート処理
 */
/* ------------------------------------------------------------------------- */
void PadState::Activate() noexcept
{
    _identify = MakeUniqueID();
    
    MGL_LIBRARY_TRACE("[MGL Gamepad] Connected: ID = %d: %s - %s: (%04X-%04X)",
                      _identify,
                      _deviceInfo.vendorName.c_str(),
                      _deviceInfo.productName.c_str(),
                      _deviceInfo.vendorID,
                      _deviceInfo.productID);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      アクティベート解除処理
 */
/* ------------------------------------------------------------------------- */
void PadState::Deactivate() noexcept
{
    MGL_LIBRARY_TRACE("[MGL Gamepad] Disconnected: ID = %d: %s - %s: (%04X-%04X)",
                      _identify,
                      _deviceInfo.vendorName.c_str(),
                      _deviceInfo.productName.c_str(),
                      _deviceInfo.vendorID,
                      _deviceInfo.productID);

    if (IsEntried())
    {
        Event::Notify(Event::NotifyType::DisconnectGamepad, this);
        GamepadServer::GetInstance().Leave(*this);
    }
    
    _identify = PadID::Invalid;
    _deviceInfo = PadDeviceInfo();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      決定とキャンセルのボタンを適用
 */
/* ------------------------------------------------------------------------- */
void PadState::ApplyDecideCancelButton() noexcept
{
    if (_decideButton != PadButton::None)
    {
        if (IsPressing(_decideButton))
        {
            SetButton(PadButton::Decide);
        }
    }
    if (_cancelButton != PadButton::None)
    {
        if (IsPressing(_cancelButton))
        {
            SetButton(PadButton::Cancel);
        }
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      アナログスティックからD-Padを適用する
 */
/* ------------------------------------------------------------------------- */
void PadState::ApplyLeftStickToDPad() noexcept
{
    constexpr struct
    {
        float angle;
        PadButtonFlags flags;
    } kTable[] =
    {
        {-2.61799f, static_cast<PadButtonFlags>(PadButton::AnalogLeft)},
        {-2.09439f, PadButton::AnalogLeft | PadButton::AnalogDown},
        {-1.04719f, static_cast<PadButtonFlags>(PadButton::AnalogDown)},
        {-0.52359f, PadButton::AnalogDown | PadButton::AnalogRight},
        { 0.52359f, static_cast<PadButtonFlags>(PadButton::AnalogRight)},
        { 1.04719f, PadButton::AnalogRight | PadButton::AnalogUp},
        { 2.09439f, static_cast<PadButtonFlags>(PadButton::AnalogUp)},
        { 2.61799f, PadButton::AnalogUp | PadButton::AnalogLeft},
        { 3.2f,     static_cast<PadButtonFlags>(PadButton::AnalogLeft)}
    };
    

    if (_leftStick.Length() >= kAnalogStickThreshold)
    {
        float angle = std::atan2f(_leftStick.y, _leftStick.x);

        for (const auto &table : kTable)
        {
            if (table.angle >= angle)
            {
                SetButton(table.flags);
                break;
            }
        }
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      リピート入力フラグを更新
 */
/* ------------------------------------------------------------------------- */
void PadState::UpdateRepeatFlags(const RepeatSetting &repeatSetting) noexcept
{
    MGL::System::Application application;

    for (size_t i = 0; i < kGamepadFlagSize; i++)
    {
        auto button = static_cast<PadButton>(1 << i);
        auto &timer = _repeatTimerArray[i];

        // 初回入力
        if (IsTriggered(button))
        {
            _repeatFlags |= 1 << i;
            timer = repeatSetting.firstBlank;
        }
        // 継続入力
        else if (IsPressing(button))
        {
            // 経過時間を更新
            if (repeatSetting.type == RepeatType::Frame)
            {
                timer -= 1.0f;
            }
            else if (repeatSetting.type == RepeatType::Second)
            {
                timer -= application.GetFrameElapsedTime() / 1000.0f;
            }

            // 指定された時間が経過していたらリピート入力をオンにする
            if (timer <= 0.0f)
            {
                _repeatFlags |= 1 << i;
                timer = repeatSetting.secondBlank;
            }
            else
            {
                _repeatFlags &= ~(1 << i);
            }
        }
        else
        {
            _repeatFlags &= ~(1 << i);
        }
    }
}

}   // MGL::Input
// vim: et ts=4 sw=4 sts=4
