// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_ascii_font.cc
 *  \brief      MGL ASCIIフォント
 *  \date       Since: May 26, 2021. 18:36:26 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/render/font/mgl_ascii_font.h>

namespace MGL::Render
{
namespace
{
    constexpr char kNullCharacter   = '\0'; // NULL文字
    constexpr char kLineFeed        = 0x0A; // 改行
    constexpr char kCarriageReturn  = 0x0D; // 行頭復帰
    constexpr char kSpace           = ' ';  // 空白
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ（テクスチャキーで初期化）
 *  \param[in]  textureKey      テクスチャキー
 *  \param[in]  characterWidth  文字の幅
 *  \param[in]  characterHeight 文字の高さ
 *  \param[in]  sourceOffset    テクスチャの参照元座標のオフセット
 */
/* ------------------------------------------------------------------------- */
AsciiFont::AsciiFont(TextureKey textureKey, uint32_t characterWidth, uint32_t characterHeight, const Vector2 &sourceOffset) noexcept
    : FontResource(false)
    , _texture(textureKey)
    , _characterSize(static_cast<float>(characterWidth), static_cast<float>(characterHeight))
    , _sourceOffset(sourceOffset)
{
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ（ファイルから画像を読み込み）
 *  \param[in]  imagePath       画像ファイルのパス
 *  \param[in]  characterWidth  文字の幅
 *  \param[in]  characterHeight 文字の高さ
 *  \param[in]  isAsync         非同期読み込みフラグ
 *  \param[in]  sourceOffset    テクスチャの参照元座標のオフセット
 *  \param[in]  loaderKey       テクスチャローダーのキー
 */
/* ------------------------------------------------------------------------- */
AsciiFont::AsciiFont(const char *imagePath, uint32_t characterWidth, uint32_t characterHeight, bool isAsync, const Vector2 &sourceOffset, TextureLoaderKey loaderKey) noexcept
    : FontResource(false)
    , _texture(imagePath, isAsync, loaderKey)
    , _characterSize(static_cast<float>(characterWidth), static_cast<float>(characterHeight))
    , _sourceOffset(sourceOffset)
{
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ（メモリ上の画像データから読み込み）
 *  \param[in]  imageData       画像データのアドレス
 *  \param[in]  dataSize        画像データのサイズ
 *  \param[in]  characterWidth  文字の幅
 *  \param[in]  characterHeight 文字の高さ
 *  \param[in]  isAsync         非同期読み込みフラグ
 *  \param[in]  sourceOffset    テクスチャの参照元座標のオフセット
 *  \param[in]  loaderKey       テクスチャローダーのキー
 */
/* ------------------------------------------------------------------------- */
AsciiFont::AsciiFont(const void *imageData, size_t dataSize, uint32_t characterWidth, uint32_t characterHeight, bool isAsync, const Vector2 &sourceOffset, TextureLoaderKey loaderKey) noexcept
    : FontResource(false)
    , _texture(imageData, dataSize, isAsync, loaderKey)
    , _characterSize(static_cast<float>(characterWidth), static_cast<float>(characterHeight))
    , _sourceOffset(sourceOffset)
{
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字の表示
 *  \param[in]  position    表示位置
 *  \param[in]  fontOption  表示オプション
 *  \param[in]  text        表示する文字列
 */
/* ------------------------------------------------------------------------- */
Vector2 AsciiFont::Print(const Vector2 position, const FontOption &fontOption, const char *text) noexcept
{
    // テクスチャが無効なら実行しない
    if (!_texture.IsValid())
    {
        return position;
    }
    
    Renderer2D renderer;
    DrawOption2D drawOption = fontOption;
    auto drawPosition = position;

    while (*text != kNullCharacter)
    {
        // 1行分表示
        text = PrintLine(drawPosition, text, renderer, drawOption, fontOption);

        // CR or CR+LFの改行の処理
        if (*text == kCarriageReturn)
        {
            text++;
            if (*text == kLineFeed)
            {
                text++;
            }
        }
        // LFの改行の処理
        else if (*text == kLineFeed)
        {
            text++;
        }

        // 改行後の表示位置の更新
        drawPosition.x = position.x;
        drawPosition.y += (_characterSize.y + fontOption.margin.y) * fontOption.scale.y;
    }

    return drawPosition;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字を1行表示
 *  \param[in]  position    表示位置
 *  \param[in]  text        表示する文字列
 *  \param[in]  renderer    使用するレンダラ
 *  \param[in]  drawOption  表示オプション
 *  \param[in]  fontOption  フォントの表示オプション
 *  \return     表示した次の文字を先頭にした文字列
 */
/* ------------------------------------------------------------------------- */
const char *AsciiFont::PrintLine(Vector2 position, const char *text, Renderer2D &renderer, const DrawOption2D &drawOption, const FontOption &fontOption) noexcept
{
    // 行の終端かをチェックするラムダ式を定義
    constexpr auto isEndOfLine = [](char character)
    {
        switch (character)
        {
            case kCarriageReturn:
            case kLineFeed:
            case kNullCharacter:
                return true;
        }
        
        return false;
    };

    // 入力された文字列の先頭が既に行末であれば何もしない
    if (isEndOfLine(*text))
    {
        return text;
    }
    
    // 水平アンカーが左寄せでなければ位置を補正
    if ((fontOption.horizontalAlignment == Alignment::Horizontal::Center) || (fontOption.horizontalAlignment == Alignment::Horizontal::Right))
    {
        // 行末までの文字数をカウントして補正するオフセット（文字幅）を計算する
        float offset = 0.0f;
        for (size_t i = 0; ; i++)
        {
            if (isEndOfLine(text[i]))
            {
                offset = static_cast<float>(i) * (_characterSize.x + fontOption.margin.x) * fontOption.scale.x;
                break;
            }
        }

        // 中央揃えの場合はオフセットを半分にする
        if (fontOption.horizontalAlignment == Alignment::Horizontal::Center)
        {
            offset *= 0.5f;
        }

        position.x -= offset;
    }

    // 行の終端まで描画
    for (; !isEndOfLine(*text); text++)
    {
        auto character = *text;

        if (character != kSpace)
        {
            Rectangle sourceRectangle(
                    (static_cast<float>(character & 0x0F) * _characterSize.x) + _sourceOffset.x,
                    (static_cast<float>(character >> 4) * _characterSize.y) + _sourceOffset.y,
                    _characterSize.x,
                    _characterSize.y
            );

            renderer.DrawSprite(position, _texture, sourceRectangle, drawOption);
        }

        position.x += (_characterSize.x + fontOption.margin.x) * fontOption.scale.x;
    }

    return text;
}

}   // namespace MGL::Render

// vim: et ts=4 sw=4 sts=4
