// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_font.cc
 *  \brief      MGL フォント
 *  \date       Since: May 27, 2021. 16:38:57 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/render/font/mgl_font.h>
#include <stdarg.h>

namespace MGL::Render
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      フォントをフォントストレージに追加
 *  \param[in]  key         関連付けるフォントキー
 *  \param[in]  resource    追加するフォントリソース
 *  \return     フォントクラス
 */
/* ------------------------------------------------------------------------- */
Font Font::AddStorage(FontKey key, SharedFontResource resource) noexcept
{
    if (!RendererSet::GetInstance().GetFontStorage().Add(key, resource))
    {
        return Font();
    }

    Font font;
    font._resource = resource;
    font._key = key;

    return font;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      フォントストレージからフォントを読み込み
 *  \param[in]  key     読み込むフォントのキー
 *  \retval     true    成功
 *  \retval     false   キーに対応したフォントが見つからなかった
 */
/* ------------------------------------------------------------------------- */
bool Font::Load(FontKey key) noexcept
{
    _resource = RendererSet::GetInstance().GetFontStorage().Get(key);
    _key = key;

    return (_resource != nullptr);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字の表示
 *  \param[in]  text    表示する文字
 *  \param[in]  args    フォーマット引数
 */
/* ------------------------------------------------------------------------- */
void Font::Print(const char *text, Text::FormatArgs args) noexcept
{
    if (_resource == nullptr)
    {
        return;
    }
    
    if (args.empty())
    {
        _position = _resource->Print(_position, _option, text);
    }
    else
    {
        _position = _resource->Print(_position, _option, MGL::Text::Format(text, args).c_str());
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      インデックス化した文字の表示
 *  \param[in]  indexedText 表示する文字
 *  \param[in]  args        フォーマット引数
 */
/* ------------------------------------------------------------------------- */
void Font::Print(const Text::IndexedCharacter *indexedText, Text::FormatArgs args) noexcept
{
    if (_resource == nullptr)
    {
        return;
    }
    
    if (args.empty())
    {
        _position = _resource->Print(_position, _option, indexedText);
    }
    else
    {
        _position = _resource->Print(_position, _option, MGL::Text::Format(indexedText, _resource->GetIndexConverter(), args).data());
    }
}

}   // namespace MGL::Render

// vim: et ts=4 sw=4 sts=4
