// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_task_node_list.cc
 *  \brief      MGL タスクノードのリスト
 *  \date       Since: June 1, 2021. 17:06:05 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/system/mgl_system_debug_macro.h>
#include <mgl/task/mgl_task_node_list.h>

namespace MGL::Task
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      インスタンスの取得
 *  \return     インスタンスの参照
 */
/* ------------------------------------------------------------------------- */
STL::unique_ptr<NodeList> &NodeList::GetInstanceRef() noexcept
{
    static STL::unique_ptr<NodeList> sInstance = nullptr;
    return sInstance;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ
 */
/* ------------------------------------------------------------------------- */
NodeList::NodeList() noexcept
    : _capacity(0)
    , _nodeElements()
    , _activeTopElement()
    , _freeTopElement()
    , _currentUniqueId(UniqueIdentifier::Start)
    , _executeSettings()
{
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      初期化処理
 *  \param[in]  capacity    タスクの最大数
 *  \param[in]  settings    実行ステージ設定
 */
/* ------------------------------------------------------------------------- */
bool NodeList::Initialize(size_t capacity, const ExecuteSettingArray &settings) noexcept
{
    // 引数チェック
    if ((capacity == 0) || (settings.empty()))
    {
        return false;
    }

    // 各メンバを初期化
    _capacity = capacity;
    _nodeElements = STL::make_unique<NodeElement[]>(capacity);
    _executeSettings = settings;

    // リストを初期化
    _freeTopElement.next = &_nodeElements[0];
    for (size_t i = 0; i < _capacity; i++)
    {
        if (i < (_capacity - 1))
        {
            _nodeElements[i].next = &_nodeElements[i + 1];
        }

        if (i == 0)
        {
            _nodeElements[i].previous = &_freeTopElement;
        }
        else
        {
            _nodeElements[i].previous = &_nodeElements[i - 1];
        }

        _nodeElements[i].arrayIndex = i;
    }
    
    // 実行設定の指定が無ければステージ0の設定を1つだけ追加
    if (_executeSettings.empty())
    {
        _executeSettings.push_back({0});
    }
    
    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ノードの追加
 *  \param[in]  addNode     追加するノード
 *  \return     追加に成功した場合はリスト上のインデックス．失敗した場合は範囲外のインデックス
 */
/* ------------------------------------------------------------------------- */
size_t NodeList::Add(STL::unique_ptr<Node> addNode) noexcept
{
    if (_freeTopElement.next == nullptr)
    {
        return _capacity;
    }

    auto *addElement = _freeTopElement.next;
    addElement->node = std::move(addNode);
    _freeTopElement.next = addElement->next;
    
    // アクティブリストが空なら先頭に挿入
    if (_activeTopElement.next == nullptr)
    {
        _activeTopElement.next = addElement;
        addElement->previous = &_activeTopElement;
        addElement->next = nullptr;
    }
    else
    {
        auto *element = _activeTopElement.next;
        while (element != nullptr)
        {
            // 追加するノードのIDよりも優先順位の低いものが見つかったらそこに挿入
            if (element->node->GetIdentifier() >= addElement->node->GetIdentifier())
            {
                addElement->previous = element->previous;
                addElement->next = element;
            }
            // 優先順位の低いノードが見つからなかったら末尾に挿入
            else if (element->next == nullptr)
            {
                addElement->previous = element;
                addElement->next = nullptr;
            }
            else
            {
                element = element->next;
                continue;
            }

            // 前後の要素のポインタを変更
            addElement->previous->next = addElement;
            if (addElement->next != nullptr)
            {
                addElement->next->previous = addElement;
            }

            break;
        }
    }
    
    // ユニークIDの設定
    addElement->uniqueId = _currentUniqueId++;
    
    // 初期化処理の呼び出し
    addElement->node->OnInitialize();

    return addElement->arrayIndex;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      削除
 *  \param[in]  element     削除するノードの要素
 */
/* ------------------------------------------------------------------------- */
void NodeList::Remove(NodeElement *element) noexcept
{
    element->node.reset();
    element->uniqueId = UniqueIdentifier::Invalid;
    
    // 削除対象の前後の要素をつなげる
    element->previous->next = element->next;
    if (element->next != nullptr)
    {
        element->next->previous = element->previous;
    }
    
    // 削除対象をフリーリストへ
    element->next = _freeTopElement.next;
    _freeTopElement.next = element;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      実行
 */
/* ------------------------------------------------------------------------- */
void NodeList::Execute() noexcept
{
    // 全てのノードのスタンバイ状態を解除
    auto *current = _activeTopElement.next;
    while (current != nullptr)
    {
        current->isStandby = false;
        current = current->next;
    }

    // ステージの数だけノードのOnExecute()を呼び出す
    for (const auto &executeSetting : _executeSettings)
    {
        current = _activeTopElement.next;
        while (current != nullptr)
        {
            auto *next = current->next;

            // 削除要求があったりスタンバイ状態でなければ実行
            if (!current->node->IsRequestedKill() || !current->isStandby)
            {
                current->node->OnExecute(executeSetting.stage);
            }

            // 削除要求があれば削除
            if (current->node->IsRequestedKill())
            {
                Remove(current);
            }

            current = next;
        }
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      全てのノードに削除要求を発行する
 *  \param[in]  resideLevel     常駐レベルがこの値以下のタスクを削除する
 *  \return     削除要求を発行したノードの数
 */
/* ------------------------------------------------------------------------- */
size_t NodeList::Kill(ResideLevel resideLevel) noexcept
{
    size_t count = 0;

    auto *current = _activeTopElement.next;
    while (current != nullptr)
    {
        if (current->node->GetResideLevel() <= resideLevel)
        {
            current->node->Kill();
            count++;
        }
        current = current->next;
    }

    return count;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      指定したIDのノードに削除要求を発行する
 *  \param[in]  identifier      削除するID
 *  \param[in]  resideLevel     常駐レベルがこの値以下のタスクを削除する
 *  \return     削除要求を発行したノードの数
 */
/* ------------------------------------------------------------------------- */
size_t NodeList::Kill(Identifier identifier, ResideLevel resideLevel) noexcept
{
    size_t count = 0;

    auto *current = _activeTopElement.next;
    while (current != nullptr)
    {
        if (identifier == current->node->GetIdentifier())
        {
            if (current->node->GetResideLevel() <= resideLevel)
            {
                current->node->Kill();
            }
        }
        else if (identifier < current->node->GetIdentifier())
        {
            break;
        }

        current = current->next;
    }

    return count;
} 


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      全体のノード数を取得
 *  \return     全体のノード数
 */
/* ------------------------------------------------------------------------- */
size_t NodeList::GetCount() noexcept
{
    size_t count = 0;

    auto *current = _activeTopElement.next;
    while (current != nullptr)
    {
        count++;
        current = current->next;
    }

    return count;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      指定したIDのノード数を取得
 *  \return     指定したIDのノード数
 */
/* ------------------------------------------------------------------------- */
size_t NodeList::GetCount(Identifier identifier) noexcept
{
    size_t count = 0;

    auto *current = _activeTopElement.next;
    while (current != nullptr)
    {
        auto currentIdentifier = current->node->GetIdentifier();

        if (identifier == currentIdentifier)
        {
            count++;
        }
        else if (identifier < currentIdentifier)
        {
            break;
        }

        current = current->next;
    }

    return count;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      リストの要素を取得
 *  \param[in]  index   要素のインデックス
 *  \return     インデックスに対応した要素．インデックスが範囲外の場合はnullptr
 */
/* ------------------------------------------------------------------------- */
const NodeList::NodeElement *NodeList::GetElement(size_t index) const noexcept
{
    if (index >= _capacity)
    {
        return nullptr;
    }
    
    return &_nodeElements[index];
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      リストの要素を取得
 *  \param[in]  node    タスクノード
 *  \return     タスクノードに対応した要素．見つからない場合はnullptr
 */
/* ------------------------------------------------------------------------- */
const NodeList::NodeElement *NodeList::GetElement(const Node *node) const noexcept
{
    auto *current = _activeTopElement.next;
    while (current != nullptr)
    {
        auto currentIdentifier = current->node->GetIdentifier();

        if (node->GetIdentifier() == currentIdentifier)
        {
            if (node == current->node.get())
            {
                return current;
            }
        }
        else if (node->GetIdentifier() < currentIdentifier)
        {
            break;
        }

        current = current->next;
    }

    return nullptr;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      IDからノードを検索
 *  \param[in]  identifier  検索するID
 *  \return     見つかったノードのインデックス
 */
/* ------------------------------------------------------------------------- */
STL::vector<size_t> NodeList::Find(Identifier identifier) const noexcept
{
    STL::vector<size_t> result;
    result.reserve((1024 - 16) / sizeof(size_t));
    
    auto *current = _activeTopElement.next;
    while (current != nullptr)
    {
        auto currentIdentifier = current->node->GetIdentifier();

        if (identifier == currentIdentifier)
        {
            result.push_back(current->arrayIndex);
        }
        else if (identifier < currentIdentifier)
        {
            break;
        }

        current = current->next;
    }
    
    return result;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      タスクへのイベント通知
 *  \param[in]  eventIdentifier     イベントID
 *  \param[in]  argument            引数
 */
/* ------------------------------------------------------------------------- */
void NodeList::NotifyEvent(EventIdentifier eventIdentifier, void *argument) noexcept
{
    auto *current = _activeTopElement.next;
    while (current != nullptr)
    {
        current->node->OnReceiveTaskEvent(eventIdentifier, argument);
        current = current->next;
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      タスクへのイベント通知
 *  \param[in]  identifier          タスクID
 *  \param[in]  eventIdentifier     イベントID
 *  \param[in]  argument            引数
 */
/* ------------------------------------------------------------------------- */
void NodeList::NotifyEvent(Identifier identifier, EventIdentifier eventIdentifier, void *argument) noexcept
{
    auto *current = _activeTopElement.next;
    while (current != nullptr)
    {
        auto currentIdentifier = current->node->GetIdentifier();

        if (identifier == currentIdentifier)
        {
            current->node->OnReceiveTaskEvent(eventIdentifier, argument);
        }
        else if (identifier < currentIdentifier)
        {
            break;
        }

        current = current->next;
    }
}

}   // namespace MGL::Task

// vim: et ts=4 sw=4 sts=4
