// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_text_character.cc
 *  \brief      文字クラス
 *  \date       Since: October 4, 2018. 2:20:42 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/text/mgl_text_character.h>
#include <string.h>

namespace MGL::Text
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字をバイトに変換する際に必要なサイズを取得
 *  \return     必要サイズ．不正な文字の場合は0
 */
/* ------------------------------------------------------------------------- */
size_t Character::GetByteSize() const noexcept
{
    if (IsValid())
    {
        // 1Byte
        if (_code <= 0x7F)
        {
            return 1;
        }
        // 2Byte
        else if (_code <= 0x7FF)
        {
            return 2;
        }
        // 3Byte
        else if (_code <= 0xFFFF)
        {
            return 3;
        }
        // 4Byte
        else if (_code <= 0x1FFFFF)
        {
            return 4;
        }
    }
    
    return 0;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字をバイト列に変換
 *  \param[out] dest        書き込み先のアドレス
 *  \param[in]  destSize    書き込み先バッファのサイズ
 *  \return     成功した場合は終端記号を書き込んだアドレス．失敗時にはnullptr
 */
/* ------------------------------------------------------------------------- */
char *Character::ToByte(char *dest, size_t destSize) const noexcept
{
    // サイズを取得して0なら失敗
    auto byteSize = GetByteSize();
    if (byteSize == 0)
    {
        return nullptr;
    }

    // 書き込み先に必要なサイズが無ければ失敗
    if (destSize <= (byteSize + 1))
    {
        return nullptr;
    }

    // サイズに応じてデコード
    switch (byteSize)
    {
        // 1Byte
        case 1:
        {
            *(dest++) = static_cast<char>(_code);
            break;
        }
            
        // 2Byte
        case 2:
        {
            uint32_t code = _code & 0x7FF;
            *(dest++) = static_cast<char>(0xC0 | (code >> 6));
            *(dest++) = static_cast<char>(0x80 | (code & 0x3F));
            break;
        }
            
        // 3Byte
        case 3:
        {
            uint32_t code = _code & 0xFFFF;
            *(dest++) = static_cast<char>(0xE0 | (code >> 12));
            *(dest++) = static_cast<char>(0x80 | ((code >> 6) & 0x3F));
            *(dest++) = static_cast<char>(0x80 | (code & 0x3F));
            break;
        }

        // 4Byte
        case 4:
        {
            uint32_t code = _code & 0x1FFFFF;
            *(dest++) = static_cast<char>(0xF0 | (code >> 18));
            *(dest++) = static_cast<char>(0x80 | ((code >> 12) & 0x3F));
            *(dest++) = static_cast<char>(0x80 | ((code >> 6) & 0x3F));
            *(dest++) = static_cast<char>(0x80 | (code & 0x3F));
            break;
        }

        // それ以外は不正な文字
        default:
            return nullptr;
    }

    // 終端コードを書き込んでそのアドレスを返す
    *dest = '\0';
    return dest;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コードをバイト列に変換
 *  \param[out] byteArray   バイトコードの格納先
 *  \retval     true        成功
 *  \retval     false       失敗
 *  \note
 *      とりあえずUTF-8で返す．
 *      将来は他の他の文字コード（Shift-JISなど）にも対応するかも．
 */
/* ------------------------------------------------------------------------- */
bool Character::ToByte(char byteArray[kByteArraySize]) const noexcept
{
    if (!IsValid())
    {
        return false;
    }

    memset(byteArray, 0, kByteArraySize);

    // 1Byte
    if (_code <= 0x7F)
    {
        byteArray[0] = static_cast<char>(_code);
    }
    // 2Byte
    else if (_code <= 0x7FF)
    {
        uint32_t code = _code & 0x7FF;
        byteArray[0] = static_cast<char>(0xC0 | (code >> 6));
        byteArray[1] = static_cast<char>(0x80 | (code & 0x3F));
    }
    // 3Byte
    else if (_code <= 0xFFFF)
    {
        uint32_t code = _code & 0xFFFF;
        byteArray[0] = static_cast<char>(0xE0 | (code >> 12));
        byteArray[1] = static_cast<char>(0x80 | ((code >> 6) & 0x3F));
        byteArray[2] = static_cast<char>(0x80 | (code & 0x3F));
    }
    // 4Byte
    else if (_code <= 0x1FFFFF)
    {
        uint32_t code = _code & 0x1FFFFF;
        byteArray[0] = static_cast<char>(0xF0 | (code >> 18));
        byteArray[1] = static_cast<char>(0x80 | ((code >> 12) & 0x3F));
        byteArray[2] = static_cast<char>(0x80 | ((code >> 6) & 0x3F));
        byteArray[3] = static_cast<char>(0x80 | (code & 0x3F));
    }
    // それ以降はUTF-8では不正なシーケンスなので失敗
    else
    {
        return false;
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字をSTL::stringに変換
 *  \return     変換した文字
 */
/* ------------------------------------------------------------------------- */
STL::string Character::ToString() const noexcept
{
    char byteArray[kByteArraySize];
    if (ToByte(byteArray))
    {
        return STL::string(byteArray);
    }

    return STL::string("");
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字列をSTL::stringに変換
 *  \return     変換した文字列
 */
/* ------------------------------------------------------------------------- */
STL::string CharacterArray::ToString() const
{
    STL::string str("");

    for (const auto &character : *this)
    {
        str += character.ToString();
    }

    return str;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字列の一部をSTL::stringに変換
 *  \param[in]  start   開始位置
 *  \param[in]  end     終了位置
 *  \return     startからendまでの文字を変換した文字列
 */
/* ------------------------------------------------------------------------- */
STL::string CharacterArray::ToString(uint32_t start, uint32_t end) const
{
    uint32_t arraySize = static_cast<uint32_t>(size());
    if ((end == 0) || (end >= arraySize))
    {
        end = arraySize - 1;
    }

    STL::string str("");

    for (uint32_t i = start; i < end; i++)
    {
        str += at(i).ToString();
    }

    return str;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字検索
 *  \param[in]  findCode    検索する文字（Unicodeスカラ値）
 *  \param[in]  start       検索開始位置
 *  \return     見つかった場合はそのインデックス，見つからない場合は負数．
 */
/* ------------------------------------------------------------------------- */
int CharacterArray::Find(uint32_t findCode, uint32_t start) const
{
    uint32_t arraySize = static_cast<uint32_t>(size());

    for (uint32_t i = start; i < arraySize; i++)
    {
        if (at(i).GetCode() == findCode)
        {
            return static_cast<int>(i);
        }
    }

    return -1;
}
}   // namespace MGL::Text

// vim: et ts=4 sw=4 sts=4

