// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_text_format_argument.cc
 *  \brief      MGL テキストフォーマットの引数
 *  \date       Since: May 28, 2021. 22:42:09 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/text/mgl_text_format_argument.h>
#include <cmath>
#include <mgl/system/mgl_system_debug_macro.h>
#include <mgl/system/mgl_system_locale.h>

namespace MGL::Text
{
namespace
{
constexpr size_t kWidthMax = 127;
constexpr size_t kPrecisionMax = 7;
}
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字列に変換
 *  \param[in]  option      変換オプション
 *  \return     変換後の文字列
 */
/* ------------------------------------------------------------------------- */
STL::string FormatArgument::ToString(const FormatArgument::Option &option) const noexcept
{
    switch(_type)
    {
        // 値
        case Type::Int:
        case Type::UnsignedInt:
        case Type::Float:
            if ((option.valueDisplayMode == Option::ValueDisplayMode::UpperCaseHex) ||
                (option.valueDisplayMode == Option::ValueDisplayMode::LowerCaseHex))
            {
                return ToHexStringFromValue(option);
            }
            else
            {
                return ToStringFromNumber(option);
            }

        // 論理値
        case Type::Bool:
            if (_b)
            {
                return AlignmentString(option, "TRUE");
            }
            return AlignmentString(option, "FALSE");

        // 文字列
        case Type::String:
            return AlignmentString(option, _string);

        // ポインタ
        case Type::Pointer:
            return ToHexStringFromValue(option);

        // それ以外は不正
        case Type::Void:
        default:
            break;
    }

    MGL_WARNING("[MGL Text] Detected invalid format argument.");
    return AlignmentString(option, "N/A");
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      フォーマット文字列のパース
 *  \param[out] option      オプションの格納先
 *  \param[in]  parseText   パースする文字列
 *  \return     パース後の先頭を指すポインタ．パース失敗時はparseTextと同じ
 */
/* ------------------------------------------------------------------------- */
const char *FormatArgument::Parse(Option &option, const char *parseText) noexcept
{
    // 文字列を数値に変換して次の文字列の先頭を取得するラムダ式を定義
    auto getNumber = [](int &number, const char *text) noexcept
    {
        STL::string str;
        while ((*text >= '0') && (*text <= '9'))
        {
            str += *text;
            text++;
        }
        
        if (!str.empty())
        {
            number = std::stoi(std::string(str.c_str()));
        }
        
        return text;
    };
    
    option = Option();
    const char *current = parseText;
    
    // 最初の文字が'{'でなければフォーマット文字列ではない
    if (*current != '{')
    {
        return parseText;
    }
    current++;
    
    // 2番目の文字列も'{'であればエスケープなのでフォーマット文字列でない
    if (*current == '{')
    {
        return current;
    }
    // 2番目の文字列が'}'であれば内容を全て省略しているのでデフォルトのオプションを返す
    else if (*current == '}')
    {
        option.isValid = true;

        return ++current;
    }
    // 2番目の文字列が数値であればインデックス指定
    else if ((*current >= '0') || (*current <= '9'))
    {
        current = getNumber(option.index, current);
        if (*current == '\0')
        {
            return parseText;
        }
    }

    // 以降はオプションの指定なので'}'が来るまで文字を解析する
    while (*current != '}')
    {
        bool skipIncrement = false;

        switch (*current)
        {
            // '}'で閉じていない場合はフォーマット文字列として扱わない
            case '\0':
                return parseText;

            // '+'は正の符号を表示
            case '+':
                option.showPlusSign = true;
                break;
                
            // ','は区切り文字を表示
            case ',':
                option.showSeparator = true;
                break;

            // '<'は左寄せ
            case '<':
                current = getNumber(option.width, ++current);
                option.width = -option.width;
                skipIncrement = true;
                break;

            // '>'は右寄せ
            case '>':
                current = getNumber(option.width, ++current);
                skipIncrement = true;
                break;

            // '.'は小数点以下の桁数の指定
            case '.':
                current = getNumber(option.precision, ++current);
                skipIncrement = true;
                break;

            // ’f’は右寄せまたは左寄せの際に埋める文字
            case 'f':
                current++;
                option.fillCharacter = *current;
                break;

            // '~'は自動インデックスをスキップ
            case '~':
                option.skipAutoIndex = true;
                break;

            // 'x'は小文字の16進数表示
            case 'x':
                current = getNumber(option.hexWidth, ++current);
                option.valueDisplayMode = Option::ValueDisplayMode::LowerCaseHex;
                skipIncrement = true;
                break;

            // 'X'は大文字の16進数表示
            case 'X':
                current = getNumber(option.hexWidth, ++current);
                option.valueDisplayMode = Option::ValueDisplayMode::UpperCaseHex;
                skipIncrement = true;
                break;

            // ':'と' 'は何もしない
            case ':':
            case ' ':
                break;

            // それ以外は不明なオプション（エラーにはせず警告ログだけ）
            default:
                MGL_WARNING("[MGL Text] Unknown format option: %c", *current);
                break;
        }
        
        if (!skipIncrement)
        {
            current++;
        }
    }

    option.isValid = true;
    
    return ++current;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      置換オプションをインデックス文字に変換
 *  \param[in]  option  オプション
 *  \return     置換オプションを表すインデックス文字
 */
/* ------------------------------------------------------------------------- */
STL::vector<IndexedCharacter> FormatArgument::ToIndexedCharacters(const FormatArgument::Option &option) noexcept
{
    STL::vector<IndexedCharacter> indices;
    indices.reserve(0xF);
    Option defaultOption;
    
    // インデックス
    if (option.index != defaultOption.index)
    {
        indices.push_back(kSubstituteFormatMaskIndex | static_cast<uint8_t>(option.index));
    }

    // 桁幅
    if (option.width != defaultOption.width)
    {
        indices.push_back(kSubstituteFormatMaskWidth | static_cast<uint8_t>(option.width));
    }

    // 小数点以下の桁幅
    if (option.precision != defaultOption.precision)
    {
        indices.push_back(kSubstituteFormatMaskPrecision | static_cast<uint8_t>(option.precision));
    }

    // 16進数の桁幅
    if (option.hexWidth != defaultOption.hexWidth)
    {
        indices.push_back(kSubstituteFormatMaskHexWidth | static_cast<uint8_t>(option.hexWidth));
    }

    // 各種フラグ
    if ((option.showPlusSign != defaultOption.showPlusSign) ||
        (option.showSeparator != defaultOption.showSeparator) ||
        (option.valueDisplayMode != defaultOption.valueDisplayMode) ||
        (option.skipAutoIndex != defaultOption.skipAutoIndex))
    {
        uint8_t flags = 0;
        
        // 符号の表示
        if (option.showPlusSign)
        {
            flags |= 1 << kSubstituteFlagBitShowPlusSign;
        }
        // 区切り文字の表示
        if (option.showSeparator)
        {
            flags |= 1 << kSubstituteFlagBitShowSeparator;
        }
        // 自動インデックスのスキップ
        if (option.skipAutoIndex)
        {
            flags |= 1 << kSubstituteFlagBitSkipAutoIndex;
        }
        
        // 値の表示モード
        flags |= static_cast<uint8_t>(option.valueDisplayMode) << kSubstituteFlagBitValueDisplayModeLo;

        indices.push_back(kSubstituteFormatMaskFlags | flags);
    }

    return indices;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      整数にしてゼロか否かを取得
 *  \retval     true    ゼロ
 *  \retval     false   ゼロではない
 */
/* ------------------------------------------------------------------------- */
bool FormatArgument::IsZero() const noexcept
{
    switch (_type)
    {
        case Type::UnsignedInt:
        case Type::Int:
            return _ui == 0;

        case Type::Float:
            return static_cast<uint32_t>(std::fabs(_f)) == 0;
            
        case Type::Bool:
            return !_b;

        case Type::String:
            return _string.empty();

        case Type::Void:
        default:
            break;
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      値が正の数かを取得
 *  \retval     true    正の数
 *  \retval     false   負の数
 */
/* ------------------------------------------------------------------------- */
bool FormatArgument::IsPositiveValue() const noexcept
{
    if (_type == Type::UnsignedInt)
    {
        return true;
    }
    else if (_type == Type::Int)
    {
        return _i >= 0;
    }
    else if (_type == Type::Float)
    {
        return _f >= 0.0f;
    }

    return false;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      値が負の数かを取得
 *  \retval     true    負の数
 *  \retval     false   正の数
 */
/* ------------------------------------------------------------------------- */
bool FormatArgument::IsNegativeValue() const noexcept
{
    if (_type == Type::Int)
    {
        return _i < 0;
    }
    else if (_type == Type::Float)
    {
        return _f < 0.0f;
    }

    return false;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      数値を文字列に変換
 *  \param[in]  option      表示オプション
 *  \return     変換後の文字列
 */
/* ------------------------------------------------------------------------- */
STL::string FormatArgument::ToStringFromNumber(const FormatArgument::Option &option) const noexcept
{
    // 値を取得
    uint64_t value = 0;
    if (_type == Type::UnsignedInt)
    {
        value = _ui;
    }
    else if (_type == Type::Int)
    {
        value = static_cast<uint64_t>((_i >= 0) ? _i : (_i * -1));
    }
    else if (_type == Type::Float)
    {
        value = static_cast<uint64_t>(std::fabs(_f));
    }

    // 区切りの桁数と記号を取得
    const auto &locale = System::Locale();
    int32_t separateNumberOfDigit = locale.GetGroupingCount();
    char separateSymbol = *(locale.GetGroupingSeparator());
    int32_t separateCount = separateNumberOfDigit;
    
    char buffer[kWidthMax + 1];
    buffer[kWidthMax] = '\0';
    auto top = &buffer[kWidthMax - 1];
    while (buffer <= top)
    {
        // 値が0なら終了
        if (value == 0)
        {
            // そもそも元の値が0
            if (IsZero())
            {
                *top = '0';                
            }
            // オプション指定がある場合，タイプが符号なし or 符号付きで正の数なら'+'を付ける
            else if (option.showPlusSign && IsPositiveValue())
            {
                *top = '+';
            }
            // 符号付きで負数なら'-'を付ける
            else if (IsNegativeValue())
            {
                *top = '-';
            }
            // それ以外
            else
            {
                top++;
            }
            
            break;
        }
        
        // オプション指定があり，区切り記号が必要な桁なら記号を付ける
        else if (option.showSeparator && (separateCount <= 0))
        {
            *top = separateSymbol;
            separateCount = separateNumberOfDigit;
        }
        // それ以外は下1桁を文字に変換
        else
        {
            separateCount--;
            *top = '0' + (value % 10);
            value /= 10;
        }

        top--;
    }

    STL::string result(top);
    
    // 小数点数の場合は小数部の部分も変換して追加
    if ((_type == Type::Float) && (option.precision != 0))
    {
        result += ToStringFromDecimal(option);
    }
    
    return AlignmentString(option, result);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          小数部の値を文字列に変換
 *  \param[in]      option      表示オプション
 */
/* ------------------------------------------------------------------------- */
STL::string FormatArgument::ToStringFromDecimal(const Option &option) const noexcept
{
    // 区切りの桁数と記号を取得
    System::Locale locale;
    int32_t separateNumberOfDigit = locale.GetGroupingCount();
    char separateSymbol = *(locale.GetGroupingSeparator());
    char pointSymbol = *(locale.GetDecimalSeparator());
    int32_t separateCount = separateNumberOfDigit;

    STL::string result;

    // 小数点を追加
    result += pointSymbol;

    // 整数部を除いた値を取得
    float f = std::fabs(_f);
    f -= std::floor(f);

    // 必要な桁数分ループ
    int max = (option.precision >= 0) ? option.precision : kPrecisionMax;
    for (int i = 0; i < max; i++)
    {
        // 小数部を10倍した値の整数部を文字に変換
        f *= 10.0f;
        float n = std::floor(f);
        result += '0' + static_cast<char>(n);
        f -= n;     // 整数部は消す

        // 桁数が指定されていない場合，値が一定以下になった時点でループを抜ける
        if ((option.precision < 0) && (f < 0.0000001f))
        {
            break;
        }

        // 区切り指定がある場合は一定桁数おきに区切り文字を挿入する
        if (option.showSeparator && (i <= (max - 2)))
        {
            if (--separateCount <= 0)
            {
                result += separateSymbol;
                separateCount = separateNumberOfDigit;
            }
        }
    }

    return result;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      値を16進数の文字列に変換
 *  \param[in]  option  オプション
 *  \return     16進数の文字列
 */
/* ------------------------------------------------------------------------- */
STL::string FormatArgument::ToHexStringFromValue(const Option &option) const noexcept
{
    constexpr const char *kUpperCaseCharacterTable = "0123456789ABCDEF";    // 大文字用テーブル
    constexpr const char *kLowerCaseCharacterTable = "0123456789abcdef";    // 小文字用テーブル

    // テーブルの選択
    const auto *characterTable = kUpperCaseCharacterTable;
    if (option.valueDisplayMode == Option::ValueDisplayMode::LowerCaseHex)
    {
        characterTable = kLowerCaseCharacterTable;
    }
    
    STL::string str;
    int count = 0;

    // 値を16進数の文字列の変換
    uint64_t value = _ui;
    while (value != 0)
    {
        auto v = (value & 0x0F);

        str = characterTable[v] + str;
        count++;
        
        value >>= 4;
    }

    // 指定した桁数だけ0を詰める
    if (option.hexWidth > count)
    {
        STL::string fill;
        auto fillCount = option.hexWidth - count;
        while (fillCount-- != 0)
        {
            fill += '0';
        }
        
        str = fill + str;
    }

    return AlignmentString(option, str);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字列のアライメント処理
 *  \param[in]  option  オプション
 *  \param[in]  string  変換元の文字列
 *  \return     変換後の文字列
 *  \note
 *      この関数はマルチバイト文字に対応していない
 */
/* ------------------------------------------------------------------------- */
STL::string FormatArgument::AlignmentString(const Option &option, const STL::string &string) const noexcept
{
    // アライメント指定が無ければ何もしない
    if (option.width == 0)
    {
        return string;
    }
    auto width = static_cast<size_t>(std::abs(option.width));

    // NOTE:
    //  length()は要素数の取得なのでマルチバイト文字を正しく処理できない．
    //  正しく対応するなら文字単位の取得とその文字が全角であるかどうかの取得が必要．
    auto length = string.length();

    // 指定された幅より長い文字列に対しては何もしない
    if (length >= width)
    {
        return string;
    }

    // 詰める文字列を生成
    STL::string fill;
    auto fillCount = width - length;
    while (fillCount-- != 0)
    {
        fill += option.fillCharacter;
    }

    // 右寄せの場合は詰める文字列を先に配置
    if (option.width > 0)
    {
        return fill + string;
    }

    // 左寄せの場合は詰める文字列を後に配置
    return string + fill;
}

}   // namespace MGL::Text

// vim: et ts=4 sw=4 sts=4
