// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_text_index_converter.cc
 *  \brief      MGL テキストのインデックス化コーンバータ
 *  \date       Since: June 27, 2021. 5:34:27 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/text/mgl_text_index_converter.h>
#include <mgl/text/mgl_text_stream.h>
#include <mgl/text/mgl_text_format_argument.h>
#include <mgl/stl/mgl_stl_containers.h>

namespace MGL::Text
{
namespace
{
constexpr char kSubstituteCode = 0x1A;   // ASCIIコードの置換
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字列をインデックス化
 *  \param[in]  text    文字列
 *  \return     インデックス化された文字列．失敗時にはnullptr
 */
/* ------------------------------------------------------------------------- */
STL::unique_ptr<IndexedCharacter[]> IndexConverter::ToIndexedString(const char *text) const noexcept
{
    if (text == nullptr)
    {
        return nullptr;
    }
    
    // 引数の置換オプションを解析
    STL::list<Text::FormatArgument::Option> formatOptions;
    STL::string substText;
    for (const char *scanText = text; *scanText != '\0'; scanText++)
    {
        Text::FormatArgument::Option formatOption;
        
        scanText = Text::FormatArgument::Parse(formatOption, scanText);
        if (formatOption.isValid)
        {
            formatOptions.push_back(formatOption);
            scanText--;
            substText += kSubstituteCode;
        }
        else
        {
            substText += *scanText;
        }
    }

    MGL::Text::TextStream stream(reinterpret_cast<const uint8_t *>(substText.c_str()), strlen(substText.c_str()), MGL::Text::Encoding::UTF8);

    // 1文字づつインデックス化
    STL::vector<MGL::Text::IndexedCharacter> indices;
    indices.reserve(strlen(text));
    for (auto character = stream.Read(); !character.IsNull(); character = stream.Read())
    {
        // 無効な文字は飛ばす
        if (!character.IsValid())
        {
            continue;
        }
        // 改行
        else if (character.IsNewLine())
        {
            indices.push_back(MGL::Text::kIndexedCharacterNewLine);
        }
        // 置換
        else if (character.GetCode() == kSubstituteCode)
        {
            if (!formatOptions.empty())
            {
                const auto &option = formatOptions.front();
                auto indexedOptions = Text::FormatArgument::ToIndexedCharacters(option);
                for (const auto index : indexedOptions)
                {
                    indices.push_back(index);
                }
                indices.push_back(MGL::Text::kIndexedCharacterSubstitute);
                formatOptions.pop_front();
            }
        }
        // それ以外はインデックスに変換
        else
        {
            MGL::Text::IndexedCharacter index = ToIndexedCharacter(character.GetCode());
            if (index == MGL::Text::kIndexedCharacterInvalid)
            {
                // TODO: 代替文字を挿入する？
                continue;
            }
            
            indices.push_back(index);
        }
    }

    indices.push_back(MGL::Text::kIndexedCharacterEndOfText);

    // ユニークポインタに変換
    auto indexedString = STL::make_unique<MGL::Text::IndexedCharacter[]>(indices.size());
    memcpy(indexedString.get(), indices.data(), sizeof(MGL::Text::IndexedCharacter) * indices.size());
    
    return indexedString;
}
}   // namespace MGL::Text

// vim: et ts=4 sw=4 sts=4
