// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_ui_event.cc
 *  \brief      MGL UIイベント
 *  \date       Since: July 7, 2021. 10:57:21 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/ui/mgl_ui_event.h>
#include <mgl/input/mouse/mgl_mouse.h>

namespace MGL::UI
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief          イベントの更新処理
 *  \param[in,out]  work            イベント用ワーク
 *  \param[in]      rectangle       判定用の矩形
 *  \param[in]      requestCancel   キャンセル要求フラグ
 *  \return         発生したイベントのタイプ
 */
/* ------------------------------------------------------------------------- */
EventType Event::Update(Work &work, const MGL::Rectangle &rectangle, bool requestCancel) noexcept
{
    EventType notifiedType = EventType::None;
    
    if (requestCancel || !IsAvailable())
    {
        if (work.state != State::Unselect)
        {
            work.state = State::Unselect;
        }
    }
    else
    {
        if (MGL::Input::Mouse().IsAvailable())
        {
            notifiedType = UpdateMouseOperation(work, rectangle);
        }
        else if (MGL::Input::Touch().IsAvailable())
        {
            notifiedType = UpdateTouchOperation(work, rectangle);
        }
    }

    return notifiedType;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          マウス操作による更新処理
 *  \param[in,out]  work            イベント用ワーク
 *  \param[in]      rectangle       判定用の矩形
 *  \return         発生したイベントのタイプ
 */
/* ------------------------------------------------------------------------- */
EventType Event::UpdateMouseOperation(Work &work, const MGL::Rectangle &rectangle) noexcept
{
    EventType notifiedType = EventType::None;
    
    MGL::Input::Mouse mouse;
    
    // マウスによる選択処理
    if (!mouse.GetDeltaMove().IsZero() || mouse.IsTriggered(MGL::Input::MouseButton::Left))
    {
        if (rectangle.IsEnclosed(mouse.GetPosition()))
        {
            if (work.state != State::Selected)
            {
                work.state = State::Selected;
                _listener->OnReceiveUIEvent(work.identifier, EventType::Selected);
                notifiedType = EventType::Selected;
            }
        }
        else if (work.state != State::Unselect)
        {
            work.state = State::Unselect;
            _listener->OnReceiveUIEvent(work.identifier, EventType::Unselect);
            notifiedType = EventType::Unselect;
        }
    }
    
    // 選択中にクリックで決定
    if (work.state == State::Selected)
    {
        if (mouse.IsTriggered(MGL::Input::MouseButton::Left))
        {
            _listener->OnReceiveUIEvent(work.identifier, EventType::Decide);
            notifiedType = EventType::Decide;
        }
    }

    return notifiedType;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          タッチ操作による更新処理
 *  \param[in,out]  work            イベント用ワーク
 *  \param[in]      rectangle       判定用の矩形
 *  \return         発生したイベントのタイプ
 */
/* ------------------------------------------------------------------------- */
EventType Event::UpdateTouchOperation(Work &work, const MGL::Rectangle &rectangle) noexcept
{
    EventType notifiedType = EventType::None;

    // 初回タッチ判定
    if (work.touchID == MGL::Input::TouchID::Invalid)
    {
        MGL::Input::Touch touch;
        
        if (touch.IsEnabled())
        {
            // 判定用矩形の内側なら選択状態にする
            if (rectangle.IsEnclosed(touch.GetPosition()))
            {
                if (work.state != State::Selected)
                {
                    work.state = State::Selected;
                    _listener->OnReceiveUIEvent(work.identifier, EventType::Selected);
                    notifiedType = EventType::Selected;
                    work.touchID = touch.GetTouchID();
                }
            }
        }
        // タッチがない状態では選択状態にはしない
        else
        {
            if (work.state != State::Unselect)
            {
                work.state = State::Unselect;
                _listener->OnReceiveUIEvent(work.identifier, EventType::Unselect);
                notifiedType = EventType::Unselect;
            }
        }
    }
    else
    {
        MGL::Input::Touch touch(work.touchID);

        // タッチ継続
        if (touch.IsEnabled())
        {
            // 矩形の内側なら選択状態
            if (rectangle.IsEnclosed(touch.GetPosition()))
            {
                if (work.state != State::Selected)
                {
                    work.state = State::Selected;
                    _listener->OnReceiveUIEvent(work.identifier, EventType::Selected);
                    notifiedType = EventType::Selected;
                }
            }
            // 矩形の外側なら非選択状態
            else
            {
                if (work.state != State::Unselect)
                {
                    work.state = State::Unselect;
                    _listener->OnReceiveUIEvent(work.identifier, EventType::Unselect);
                    notifiedType = EventType::Unselect;
                }
            }
        }
        // タッチ終了
        else
        {
            // 終了した時点で選択状態であれば決定
            if (work.state == State::Selected)
            {
                _listener->OnReceiveUIEvent(work.identifier, EventType::Decide);
                notifiedType = EventType::Decide;
            }
            
            work.touchID = MGL::Input::TouchID::Invalid;
        }
    }
    
    return notifiedType;
}

}   // namespace UI

// vim: et ts=4 sw=4 sts=4
