// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_ui_widget.cc
 *  \brief      MGL UIウィジット
 *  \date       Since: July 5, 2021. 8:12:05 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/ui/mgl_ui_widget.h>
#include <mgl/render/mgl_renderer_2d.h>

namespace MGL::UI
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      更新処理
 *  \param[in]  offset      オフセット
 *  \param[in]  updateEvent イベント更新フラグ
 */
/* ------------------------------------------------------------------------- */
void Widget::Update(const MGL::Vector2 &offset, bool updateEvent) noexcept
{
    OnUpdate(GetAdjustedPosition() + offset);
    
    auto updatedPosition(GetAdjustedPosition() + offset);
    for (auto &child : _children)
    {
        child->Update(updatedPosition, false);
    }
    
    if (updateEvent)
    {
        bool cancelRequest = false;
        UpdateEvent(offset, cancelRequest);
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          イベントの更新処理
 *  \param[in]      offset          オフセット
 *  \param[in,out]  cancelRequest   キャンセル要求
 */
/* ------------------------------------------------------------------------- */
void Widget::UpdateEvent(const MGL::Vector2 &offset, bool &cancelRequest) noexcept
{
    auto position(GetAdjustedPosition() + offset);
    
    for (auto it = _children.rbegin(); it != _children.rend(); ++it)
    {
        auto *child = *it;
        child->UpdateEvent(position, cancelRequest);
    }
    
    if (_event != nullptr)
    {
        EventType notifiedType = EventType::None;
        if (IsVisible() && _isEnabledEvent)
        {
            MGL::Rectangle rectangle(GetAdjustedPosition() + offset, GetRectangle().GetSize());
            notifiedType = _event->Update(_eventWork, rectangle, cancelRequest);
        }
        else if (_eventWork.state == Event::State::Selected)
        {
            notifiedType = _event->Update(_eventWork, MGL::Rectangle(), true);
        }

        if (notifiedType != EventType::None)
        {
            OnNotifiedEvent(notifiedType);
        }
        if (_eventWork.state == Event::State::Selected)
        {
            cancelRequest = true;
        }
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      描画処理
 *  \param[in]  offset  オフセット
 */
/* ------------------------------------------------------------------------- */
void Widget::Render(const MGL::Vector2 &offset) noexcept
{
    if (!IsVisible())
    {
        return;
    }

    OnRender(GetAdjustedPosition() + offset);

    auto updatedPosition(GetAdjustedPosition() + offset);
    for (auto &child : _children)
    {
        child->Render(updatedPosition);
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      可視状態の取得
 *  \retval     true        可視
 *  \retval     false       不可視
 */
/* ------------------------------------------------------------------------- */
bool Widget::IsVisible() const noexcept
{
    auto *widget = this;
    while (widget != nullptr)
    {
        if (!widget->_isVisible)
        {
            return false;
        }
        widget = widget->_parent;
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      透過値の取得
 *  \return     透過値
 */
/* ------------------------------------------------------------------------- */
float Widget::GetTransparency() const noexcept
{
    float value = _transparency;

    auto *parent = _parent;
    while (parent != nullptr)
    {
        value *= parent->_transparency;
        parent = parent->_parent;
    }

    return value;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      親のサイズを取得
 *  \return     親のサイズ．親がいない場合はレンダーターゲットのサイズ
 */
/* ------------------------------------------------------------------------- */
MGL::Vector2 Widget::GetParentSize() const noexcept
{
    if (_parent == nullptr)
    {
        return MGL::Render::Renderer2D().GetRenderTarget().GetSize();
    }
    
    return _parent->GetRectangle().GetSize();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      位置を取得
 *  \return     ウィジットの位置
 */
/* ------------------------------------------------------------------------- */
MGL::Vector2 Widget::GetPosition() const noexcept
{
    auto position = GetAdjustedPosition();
    
    auto *widget = _parent;
    while (widget != nullptr)
    {
        position += widget->GetAdjustedPosition();
        widget = widget->_parent;
    }
    
    return position;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      子ウィジットの追加
 *  \param[in]  child   子ウィジット
 */
/* ------------------------------------------------------------------------- */
bool Widget::AddChild(Widget *child) noexcept
{
    if (child == nullptr)
    {
        return false;
    }
    
    _children.push_back(child);

    child->_parent = this;
    
    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      子ウィジットの削除
 *  \param[in]  child   削除するウィジット
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
bool Widget::RemoveChild(Widget *child) noexcept
{
    if (child == nullptr)
    {
        return false;
    }

    for (auto it = _children.begin(); it != _children.end(); ++it)
    {
        if (*it == child)
        {
            _children.erase(it);
            return true;
        }
    }

    return false;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      位置情報によって補正された座標を取得
 *  \return     補正後の座標
 */
/* ------------------------------------------------------------------------- */
MGL::Vector2 Widget::GetAdjustedPosition() const noexcept
{
    MGL::Vector2 parentSize;

    auto parent = _parent;
    while (parent != nullptr)
    {
        parentSize = parent->GetRectangle().GetSize();
        if (!parentSize.IsZero())
        {
            break;
        }
        
        parent = parent->_parent;
    }
    if (parentSize.IsZero())
    {
        parentSize = MGL::Render::Renderer2D().GetRenderTarget().GetSize();
    }
    
    auto adjustedPosition = _pivotAlignment.AdjustRectangle(GetRectangle());
    adjustedPosition = _anchorAlignment.AdjustPosition(parentSize, adjustedPosition);
    
    return adjustedPosition;
}
}   // namespace UI

// vim: et ts=4 sw=4 sts=4
